"""Widget to select the basemap to visualise on a ipyleaflet Map"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyleaflet
from ipyleaflet import basemaps as ipybasemaps, basemap_to_tiles, LayerGroup

try:
    from jeodpp import inter
except:
    pass

import traitlets

try:
    from . import settings
    from . import treeview
except:
    import settings
    import treeview


# Empty basemap (gray background)
emptyBasemap = traitlets.Bunch({'attribution': '',
                                'max_zoom': 19, 
                                'name': 'Empty basemap',
                                'build_url':   lambda *args, **kwargs: 'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/gray.png'})



# Additional attributions for the basemaps not already in ipyleaflet
basemapsAttribution = {
    'Wikimedia':                    'Wikimedia maps beta | &copy; <a href="http://osm.org/copyright" target="_blank">OpenStreetMap</a> contributors', 
    'CartoDB.PositronOnlyLabels':   '&copy; <a href="http://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions" target="_blank">CartoDB</a>',
    'CartoDB.DarkMatterOnlyLabels': '&copy; <a href="http://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions" target="_blank">CartoDB</a>',
    'CartoDB.VoyagerOnlyLabels':    '&copy; <a href="http://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions" target="_blank">CartoDB</a>',
    'OpenStreetMap.EC':             '&copy; <a href="http://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a>',
    'Google.Roadmap':               '&copy; <a href="http://www.google.com" target="_blank">Google</a>',
    'Google.Satellite':             '&copy; <a href="http://www.google.com" target="_blank">Google</a>',
    'Google.Terrain':               '&copy; <a href="http://www.google.com" target="_blank">Google</a>',
    'Google.Hybrid':                '&copy; <a href="http://www.google.com" target="_blank">Google</a>',
    'Esri.Ocean':                   'Tiles &copy; Esri — GEBCO, NOAA, National Geographic, DeLorme, HERE, Geonames.org, and other contributors', 
    'Esri.Terrain':                 'Tiles &copy; Esri — Esri, USGS, NOAA',
    'Esri.ShadedRelief':            'Tiles &copy; Esri',
    'Esri.PhysicalMap':             'Tiles &copy; Esri — U.S. National Park Service',
    'Stamen.TonerBackground':       'Map tiles by <a href=\"http://stamen.com\" target="_blank">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\" target="_blank">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a>', 
    'Stamen.TonerLite':             'Map tiles by <a href=\"http://stamen.com\" target="_blank">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\" target="_blank">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a>', 
    'Stamen.TerrainBackground':     'Map tiles by <a href=\"http://stamen.com\" target="_blank">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\" target="_blank">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a>', 
    'CyclOSM':                      '<a href=\"https://github.com/cyclosm/cyclosm-cartocss-style/releases\" title=\"CyclOSM - Open Bicycle render\" target="_blank">CyclOSM</a> | Map data: &copy; <a href=\"https://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a> contributors', 
    'CartoDB.PositronNoLabels':     '&copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\" target="_blank">CartoDB</a>', 
    'CartoDB.DarkMatterNoLabels':   '&copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\" target="_blank">CartoDB</a>', 
    'CartoDB.Voyager':              '&copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\" target="_blank">CartoDB</a>', 
    'CartoDB.VoyagerNoLabels':      '&copy; <a href=\"http://www.openstreetmap.org/copyright\" target="_blank">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\" target="_blank">CartoDB</a>', 
    'Gisco.OSMCartoComposite':      '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoBackground':     '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoLabels':         '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBlossomComposite':    '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBlossomBackground':   '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBlossomLabels':       '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMPositronComposite':   '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMPositronBackground':  '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMPositronLabels':      '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoV4Composite':    '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoV4Background':   '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoV4Labels':       '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBrightComposite':     '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBrightBackground':    '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMBrightLabels':        '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoHDMComposite':   '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoHDMBackground':  '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
    'Gisco.OSMCartoHDMLabels':      '&copy; <a href=\"https://ec.europa.eu/eurostat/web/gisco\" target="_blank">Eurostat GISCO</a>', 
  
    'BDAP.Elevation.Merit':                  '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — Yamazaki D. et al.: A high accuracy map of global terrain elevations',
    'BDAP.Elevation.Gebco':                  '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — <a href=\"https://www.gebco.net/data_and_products/gebco_web_services/web_map_service/\" target="_blank">General Bathymetric Chart of the Oceans</a>',
    'BDAP.GridSystems.MGRS':                 '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — US Army Map Service (now NGA - National Geospatial-Intelligence Agency',
    'BDAP.GridSystems.Landsat':              '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — USGS',
    'BDAP.GridSystems.UTMgrid':              '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — ESRI',
    'BDAP.Orthoimagery.Terracolor':          '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — Earthstar Geographics LLC',
    'BDAP.GlobalForestChange.LandsatMosaic': '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — Hansen, UMD, Google, USGS, NASA',
    'BDAP.GHSL.S2Mosaic':                    '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a> — GHSL',
    'BDAP.Core003.SpotMosaic':               '&copy; Tiles served by <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\" target="_blank">EC-JRC BDAP</a>',
}


# Build a basemap that works in old and new versions of ipyleaflet
def buildBasemap(oldstyledictionary):
    if ipyleaflet.__version__ == '0.9.0':
        return oldstyledictionary
    else:
        return traitlets.Bunch({'max_zoom':    oldstyledictionary['max_zoom'],
                                'name':        oldstyledictionary['name'],
                                'attribution': oldstyledictionary['attribution'],
                                'build_url':   lambda *args, **kwargs: oldstyledictionary['url']})

    
# Given a name and an interapro collections path (example inter.collections.BaseData.Elevation.MERIT.Hillshade), returns a basemap
def buildBdapLayer(name, collectionpath):
    pBack = inter.Collection(collectionpath).process()
    procid = pBack.toLayer()
    return buildBasemap({'attribution': '&copy; <a href=\"https://jeodpp.jrc.ec.europa.eu/bdap/\">EC-JRC BDAP</a>', 
                         'max_zoom': 18,
                         'name': name,
                         'url': 'https://jeodpp.jrc.ec.europa.eu/jeodpp-inter-view/?x={x}&y={y}&z={z}&procid=%s'%procid})
    

# Given a int or a string returns a basemap
def preprocessBasemap(bm):
    if type(bm) is int or type(bm) is str:
        if bm == 1 or bm == 'OpenStreetMap.Mapnik':
            return ipybasemaps.OpenStreetMap.Mapnik
        elif bm == 2 or bm == 'OpenStreetMap.BlackAndWhite':
            return ipybasemaps.OpenStreetMap.BlackAndWhite
        elif bm == 3 or bm == 'OpenStreetMap.DE':
            return ipybasemaps.OpenStreetMap.DE
        elif bm == 4 or bm == 'OpenStreetMap.France':
            return ipybasemaps.OpenStreetMap.France
        elif bm == 5 or bm == 'OpenStreetMap.HOT':
            return ipybasemaps.OpenStreetMap.HOT
        elif bm == 6 or bm == 'OpenTopoMap':
            return ipybasemaps.OpenTopoMap
        elif bm == 7 or bm == 'OpenMapSurfer.Roads':
            return ipybasemaps.OpenMapSurfer.Roads
        elif bm == 8 or bm == 'OpenMapSurfer.Grayscale':
            return ipybasemaps.OpenMapSurfer.Grayscale
        elif bm == 9 or bm == 'Hydda.Full':
            return ipybasemaps.Hydda.Full
        elif bm == 10 or bm == 'Hydda.Base':
            return ipybasemaps.Hydda.Base
        elif bm == 11 or bm == 'Esri.WorldStreetMap':
            return ipybasemaps.Esri.WorldStreetMap
        elif bm == 12 or bm == 'Esri.DeLorme':
            return ipybasemaps.Esri.DeLorme
        elif bm == 13 or bm == 'Esri.WorldTopoMap':
            return ipybasemaps.Esri.WorldTopoMap
        elif bm == 14 or bm == 'Esri.WorldImagery':
            return ipybasemaps.Esri.WorldImagery
        elif bm == 15 or bm == 'Esri.NatGeoWorldMap':
            return ipybasemaps.Esri.NatGeoWorldMap
        elif bm == 16 or bm == 'HikeBike.HikeBike':
            return ipybasemaps.HikeBike.HikeBike
        elif bm == 17 or bm == 'Stamen.Terrain':
            return ipybasemaps.Stamen.Terrain
        elif bm == 18 or bm == 'MtbMap':
            return ipybasemaps.MtbMap
        elif bm == 19 or bm == 'CartoDB.Positron':
            return ipybasemaps.CartoDB.Positron
        elif bm == 20 or bm == 'CartoDB.DarkMatter':
            return ipybasemaps.CartoDB.DarkMatter
        elif bm == 21 or bm == 'NASAGIBS.ModisTerraTrueColorCR':
            return ipybasemaps.NASAGIBS.ModisTerraTrueColorCR
        elif bm == 22 or bm == 'NASAGIBS.ModisTerraBands367CR':
            return ipybasemaps.NASAGIBS.ModisTerraBands367CR
        elif bm == 23 or bm == 'NASAGIBS.ModisTerraBands721CR':
            return ipybasemaps.NASAGIBS.ModisTerraBands721CR
        elif bm == 24 or bm == 'NASAGIBS.ModisAquaTrueColorCR':
            return ipybasemaps.NASAGIBS.ModisAquaTrueColorCR
        elif bm == 25 or bm == 'NASAGIBS.ModisAquaBands721CR':
            return ipybasemaps.NASAGIBS.ModisAquaBands721CR
        elif bm == 26 or bm == 'NASAGIBS.ViirsEarthAtNight2012':
            return ipybasemaps.NASAGIBS.ViirsEarthAtNight2012
        elif bm == 27 or bm == 'Strava.All':
            return ipybasemaps.Strava.All
        elif bm == 28 or bm == 'Strava.Ride':
            return ipybasemaps.Strava.Ride
        elif bm == 29 or bm == 'Strava.Run':
            return ipybasemaps.Strava.Run
        elif bm == 30 or bm == 'Strava.Water':
            return ipybasemaps.Strava.Water
        elif bm == 31 or bm == 'Strava.Winter':
            return ipybasemaps.Strava.Winter
        elif bm == 33 or bm == 'Stamen.Toner':
            return ipybasemaps.Stamen.Toner
        elif bm == 34 or bm == 'Stamen.Watercolor':
            return ipybasemaps.Stamen.Watercolor
        elif bm == 35 or bm == 'Wikimedia':
            #return buildBasemap({'attribution': 'Wikimedia maps beta | &copy; <a href="http://osm.org/copyright">OpenStreetMap</a> contributors', 
            #                     'max_zoom': 19,
            #                     'name': 'Wikimedia',
            #                     'url': 'https://maps.wikimedia.org/osm-intl/{z}/{x}/{y}@2x.png'})
            return buildBdapLayer(bm, inter.collections.Basemaps.Wikimedia)  # It works as a BDAP layer!
        elif bm == 36 or bm == 'CartoDB.PositronOnlyLabels':
            return buildBasemap({'attribution': '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions">CartoDB</a>',
                                 'max_zoom': 19, 'name':
                                 'CartoDB.PositronOnlyLabels',
                                 'url': 'https://cartodb-basemaps-a.global.ssl.fastly.net/rastertiles/light_only_labels/{z}/{x}/{y}.png'})
        elif bm == 37 or bm == 'CartoDB.DarkMatterOnlyLabels':
            return buildBasemap({'attribution': '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions">CartoDB</a>',
                                 'max_zoom': 19,
                                 'name': 'CartoDB.DarkMatterOnlyLabels',
                                 'url': 'https://cartodb-basemaps-a.global.ssl.fastly.net/rastertiles/dark_only_labels/{z}/{x}/{y}.png'})
        elif bm == 38 or bm == 'CartoDB.VoyagerOnlyLabels':
            return buildBasemap({'attribution': '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions">CartoDB</a>',
                                 'max_zoom': 19,
                                 'name': 'CartoDB.VoyagerOnlyLabels',
                                 'url': 'https://cartodb-basemaps-a.global.ssl.fastly.net/rastertiles/voyager_only_labels/{z}/{x}/{y}.png'})
        elif bm == 39 or bm == 'OpenStreetMap.EC':
            return buildBasemap({'attribution': '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
                                 'max_zoom': 18,
                                 'name': 'OpenStreetMap.EC',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 40 or bm == 'Google.Roadmap':
            return buildBasemap({'attribution': '&copy; <a href="http://www.google.com">Google</a>',
                                 'max_zoom': 18,
                                 'name': 'Google.Roadmap',
                                 'url': 'https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}'})
        elif bm == 41 or bm == 'Google.Satellite':
            return buildBasemap({'attribution': '&copy; <a href="http://www.google.com">Google</a>',
                                 'max_zoom': 18,
                                 'name': 'Google.Satellite',
                                 'url': 'https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}'})
        elif bm == 42 or bm == 'Google.Terrain':
            return buildBasemap({'attribution': '&copy; <a href="http://www.google.com">Google</a>',
                                 'max_zoom': 18,
                                 'name': 'Google.Terrain',
                                 'url': 'https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}'})
        elif bm == 43 or bm == 'Google.Hybrid':
            return buildBasemap({'attribution': '&copy; <a href="http://www.google.com">Google</a>',
                                 'max_zoom': 18,
                                 'name': 'Google.Hybrid',
                                 'url': 'https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}'})
        elif bm == 44 or bm == 'Esri.Ocean':
            return buildBasemap({'attribution': '&copy; <a href="http://www.esri.com">Esri</a>',
                                 'max_zoom': 18,
                                 'name': 'Esri.Ocean',
                                 'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Ocean/World_Ocean_Base/MapServer/tile/{z}/{y}/{x}'})
        elif bm == 45 or bm == 'Esri.Terrain':
            return buildBasemap({'attribution': '&copy; <a href="http://www.esri.com">Esri</a>',
                                 'max_zoom': 18,
                                 'name': 'Esri.Terrain',
                                 'url': 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer/tile/{z}/{y}/{x}'})
        elif bm == 46 or bm == 'Esri.ShadedRelief':
            return buildBasemap({'attribution': '&copy; <a href="http://www.esri.com">Esri</a>',
                                 'max_zoom': 18,
                                 'name': 'Esri.ShadedRelief',
                                 'url': 'https://services.arcgisonline.com/arcgis/rest/services/World_Shaded_Relief/MapServer/tile/{z}/{y}/{x}'})
        elif bm == 47 or bm == 'Esri.PhysicalMap':
            return buildBasemap({'attribution': '&copy; <a href="http://www.esri.com">Esri</a>', 
                                 'max_zoom': 18,
                                 'name': 'Esri.PhysicalMap',
                                 'url': 'https://services.arcgisonline.com/arcgis/rest/services/World_Physical_Map/MapServer/tile/{z}/{y}/{x}'})
        elif bm == 48 or bm == 'Stamen.TonerBackground':
            return buildBasemap({'attribution': 'Map tiles by <a href=\"http://stamen.com\">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a>', 
                                 'max_zoom': 18,
                                 'name': 'Stamen.TonerBackground',
                                 'url': 'https://stamen-tiles-a.a.ssl.fastly.net/toner-background/{z}/{x}/{y}.png'})
        elif bm == 49 or bm == 'Stamen.TonerLite':
            return buildBasemap({'attribution': 'Map tiles by <a href=\"http://stamen.com\">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a>', 
                                 'max_zoom': 18,
                                 'name': 'Stamen.TonerLite',
                                 'url': 'https://stamen-tiles-a.a.ssl.fastly.net/toner-lite/{z}/{x}/{y}.png'})
        elif bm == 50 or bm == 'Stamen.TerrainBackground':
            return buildBasemap({'attribution': 'Map tiles by <a href=\"http://stamen.com\">Stamen Design</a>, <a href=\"http://creativecommons.org/licenses/by/3.0\">CC BY 3.0</a> &mdash; Map data &copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a>', 
                                 'max_zoom': 18,
                                 'name': 'Stamen.TerrainBackground',
                                 'url': 'https://stamen-tiles-a.a.ssl.fastly.net/terrain-background/{z}/{x}/{y}.png'})
        elif bm == 51 or bm == 'CyclOSM':
            return buildBasemap({'attribution': '<a href=\"https://github.com/cyclosm/cyclosm-cartocss-style/releases\" title=\"CyclOSM - Open Bicycle render\">CyclOSM</a> | Map data: &copy; <a href=\"https://www.openstreetmap.org/copyright\">OpenStreetMap</a> contributors', 
                                 'max_zoom': 18,
                                 'name': 'CyclOSM',
                                 'url': 'https://a.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png'})
        elif bm == 52 or bm == 'Stadia.AlidadeSmooth':
            return buildBasemap({'attribution': '&copy; <a href=\"https://stadiamaps.com/\">Stadia Maps</a>, &copy; <a href=\"https://openmaptiles.org/\">OpenMapTiles</a> &copy; <a href=\"http://openstreetmap.org\">OpenStreetMap</a> contributors', 
                                 'max_zoom': 18,
                                 'name': 'Stadia.AlidadeSmooth',
                                 'url': 'https://tiles.stadiamaps.com/tiles/alidade_smooth/{z}/{x}/{y}'})
        elif bm == 53 or bm == 'Stadia.AlidadeSmoothDark':
            return buildBasemap({'attribution': '&copy; <a href=\"https://stadiamaps.com/\">Stadia Maps</a>, &copy; <a href=\"https://openmaptiles.org/\">OpenMapTiles</a> &copy; <a href=\"http://openstreetmap.org\">OpenStreetMap</a> contributors', 
                                 'max_zoom': 18,
                                 'name': 'Stadia.AlidadeSmoothDark',
                                 'url': 'https://tiles.stadiamaps.com/tiles/alidade_smooth_dark/{z}/{x}/{y}'})
        elif bm == 54 or bm == 'Stadia.OSMBright':
            return buildBasemap({'attribution': '&copy; <a href=\"https://stadiamaps.com/\">Stadia Maps</a>, &copy; <a href=\"https://openmaptiles.org/\">OpenMapTiles</a> &copy; <a href=\"http://openstreetmap.org\">OpenStreetMap</a> contributors', 
                                 'max_zoom': 18,
                                 'name': 'Stadia.OSMBright',
                                 'url': 'https://tiles.stadiamaps.com/tiles/osm_bright/{z}/{x}/{y}'})
        elif bm == 55 or bm == 'Stadia.Outdoors':
            return buildBasemap({'attribution': '&copy; <a href=\"https://stadiamaps.com/\">Stadia Maps</a>, &copy; <a href=\"https://openmaptiles.org/\">OpenMapTiles</a> &copy; <a href=\"http://openstreetmap.org\">OpenStreetMap</a> contributors', 
                                 'max_zoom': 18,
                                 'name': 'Stadia.Outdoors',
                                 'url': 'https://tiles.stadiamaps.com/tiles/outdoors/{z}/{x}/{y}'})
        elif bm == 56 or bm == 'CartoDB.PositronNoLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\">CartoDB</a>', 
                                 'max_zoom': 18,
                                 'name': 'CartoDB.PositronNoLabels',
                                 'url': 'https://a.basemaps.cartocdn.com/light_nolabels/{z}/{x}/{y}.png'})
        elif bm == 57 or bm == 'CartoDB.DarkMatterNoLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\">CartoDB</a>', 
                                 'max_zoom': 18,
                                 'name': 'CartoDB.DarkMatterNoLabels',
                                 'url': 'https://a.basemaps.cartocdn.com/dark_nolabels/{z}/{x}/{y}.png'})
        elif bm == 58 or bm == 'CartoDB.Voyager':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\">CartoDB</a>', 
                                 'max_zoom': 18,
                                 'name': 'CartoDB.Voyager',
                                 'url': 'https://a.basemaps.cartocdn.com/rastertiles/voyager/{z}/{x}/{y}.png'})
        elif bm == 59 or bm == 'CartoDB.VoyagerNoLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &copy; <a href=\"http://cartodb.com/attributions\">CartoDB</a>', 
                                 'max_zoom': 18,
                                 'name': 'CartoDB.VoyagerNoLabels',
                                 'url': 'https://a.basemaps.cartocdn.com/rastertiles/voyager_nolabels/{z}/{x}/{y}.png'})
        elif bm == 60 or bm == 'Gisco.OSMCartoComposite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoComposite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 61 or bm == 'Gisco.OSMCartoBackground':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoBackground',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoBackground/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 62 or bm == 'Gisco.OSMCartoLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoLabels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoLabels/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 63 or bm == 'Gisco.OSMBlossomComposite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBlossomComposite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBlossomComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 64 or bm == 'Gisco.OSMBlossomBackground':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBlossomBackground',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBlossomBackground/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 65 or bm == 'Gisco.OSMBlossomLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBlossomLabels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBlossomLabels/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 66 or bm == 'Gisco.OSMPositronComposite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMPositronComposite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMPositronComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 67 or bm == 'Gisco.OSMPositronBackground':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMPositronBackground',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMPositronBackground/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 68 or bm == 'Gisco.OSMPositronLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMPositronLabels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMPositronLabels/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 69 or bm == 'Gisco.OSMCartoV4Composite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoV4Composite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoV4Composite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 70 or bm == 'Gisco.OSMCartoV4Background':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoV4Background',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoV4Background/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 71 or bm == 'Gisco.OSMCartoV4Labels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoV4Labels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoV4Labels/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 72 or bm == 'Gisco.OSMBrightComposite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBrightComposite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBrightComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 73 or bm == 'Gisco.OSMBrightBackground':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBrightBackground',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBrightBackground/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 74 or bm == 'Gisco.OSMBrightLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMBrightLabels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMBrightLabels/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 75 or bm == 'Gisco.OSMCartoHDMComposite':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoHDMComposite',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoHDMComposite/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 76 or bm == 'Gisco.OSMCartoHDMBackground':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoHDMBackground',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoHDMBackground/EPSG3857/{z}/{x}/{y}.png'})
        elif bm == 77 or bm == 'Gisco.OSMCartoHDMLabels':
            return buildBasemap({'attribution': '&copy; <a href=\"http://www.openstreetmap.org/copyright\">OpenStreetMap</a> &amp; ESTAT', 
                                 'max_zoom': 18,
                                 'name': 'Gisco.OSMCartoHDMLabels',
                                 'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoHDMLabels/EPSG3857/{z}/{x}/{y}.png'})
        
        # Additional basemaps from BDAP
        elif bm == 'BDAP.Elevation.Merit':
            return buildBdapLayer(bm, inter.collections.BaseData.Elevation.MERIT.Hillshade)
        elif bm == 'BDAP.Elevation.Gebco':
            return buildBdapLayer(bm, inter.collections.BaseData.Elevation.GEBCO.Hillshade)
        elif bm == 'BDAP.GridSystems.MGRS':
            return buildBdapLayer(bm, inter.collections.BaseData.GeographicalGridSystems.MGRS)
        elif bm == 'BDAP.GridSystems.Landsat':
            return buildBdapLayer(bm, inter.collections.BaseData.GeographicalGridSystems.Landsat_WRS2_Descending)
        elif bm == 'BDAP.GridSystems.UTMgrid':
            return buildBdapLayer(bm, inter.collections.BaseData.GeographicalGridSystems.UTMgrid)
        elif bm == 'BDAP.Orthoimagery.Terracolor':
            return buildBdapLayer(bm, inter.collections.BaseData.Orthoimagery.Terracolor)
        elif bm == 'BDAP.GlobalForestChange.LandsatMosaic':
            return buildBdapLayer(bm, inter.collections.BaseData.Landcover.GFC_UMD.Version_1_7.last)
        elif bm == 'BDAP.GHSL.S2Mosaic':
            return buildBdapLayer(bm, inter.collections.Products.Mosaics.Global.GHS_composite_S2_L1C_2017_2018_GLOBE_R2020A)
        elif bm == 'BDAP.Core003.SpotMosaic':
            return buildBdapLayer(bm, inter.collections.Products.Mosaics.Europe.Core003.Seamline)

    return bm


# Retuns the list of all the basemap names
def basemapList(addBDAPbasemaps=True, removeBasemaps=[]):
    c = inter.ImageCollection("BASEMAP")
    names = list(c.listBasemaps())
    
    # Remove basemaps not working
    names.remove('OpenStreetMap.BlackAndWhite')
    names.remove('HikeBike.HikeBike')
    names.remove('Hydda.Base')
    names.remove('Hydda.Full')
    names.remove('OpenRailwayMap')
    names.remove('OpenMapSurfer.Grayscale')
    names.remove('OpenMapSurfer.Roads')
    names.remove('Stadia.AlidadeSmooth')
    names.remove('Stadia.AlidadeSmoothDark')
    names.remove('Stadia.OSMBright')
    names.remove('Stadia.Outdoors')
    
    # Remove basemaps not wanted by the caller
    for bm in removeBasemaps:
        if bm in names:
            names.remove(bm)
    
    # Add BDAP basemaps
    if addBDAPbasemaps:
        names.append('BDAP.Elevation.Merit')
        names.append('BDAP.Elevation.Gebco')
        names.append('BDAP.GridSystems.MGRS')
        names.append('BDAP.GridSystems.Landsat')
        names.append('BDAP.GridSystems.UTMgrid')
        names.append('BDAP.Orthoimagery.Terracolor')
        names.append('BDAP.GlobalForestChange.LandsatMosaic')
        names.append('BDAP.GHSL.S2Mosaic')
        names.append('BDAP.Core003.SpotMosaic')
        
    names = sorted(names)
    return names


# Assign an attribution to a TileLayer if it doesn't already have it
def assignAttribution(bm, tilelayer):
    if len(tilelayer.attribution) <= 0: 
        if bm in basemapsAttribution:
            tilelayer.attribution = basemapsAttribution[bm]


# Given an int, a string or a basemaps object returns a tile layer to add to an ipyleaflet map
def basemapTileLayer(bm):
    base = preprocessBasemap(bm)

    # basemaps that need a background
    if bm in ['BDAP.Elevation.Merit', 'BDAP.GridSystems.MGRS', 'BDAP.GridSystems.Landsat', 'BDAP.GridSystems.UTMgrid', 
              'BDAP.Orthoimagery.Terracolor', 'BDAP.GHSL.S2Mosaic', 'BDAP.Core003.SpotMosaic']:

        backbm = 'Gisco.OSMCartoComposite'
        back = preprocessBasemap(backbm)
        tileback = basemap_to_tiles(back)
        assignAttribution(backbm,tileback)
        
        tilebase = basemap_to_tiles(base)
        assignAttribution(bm,tilebase)
        tilelayer = LayerGroup(name=bm,layers=(tileback,tilebase))
    else:
        tilelayer = basemap_to_tiles(base)
        assignAttribution(bm,tilelayer)

    tilelayer.base = True
    return tilelayer



# Clear the map
def map_clear(m):
    emptyLayer = basemap_to_tiles(emptyBasemap)
    emptyLayer.base = True
    
    m.clear()
    m.add_layer(emptyLayer)
    m.layers = (m.layers[1],)
    
    
# Change the basemap for a map and returns the layer
def map_setbasemap(m, bm=None):
    if bm is None:
        layer = basemap_to_tiles(emptyBasemap)
        layer.base = True
    else:
        layer = basemapTileLayer(bm)
    newlayers = [layer] + list(m.layers[1:])
    m.layers = tuple(newlayers)
   
    return layer

    
#####################################################################################################################################################
# Basemaps selection widget class
#####################################################################################################################################################
class basemaps():
    """
    Treeview widget to select a basemap for an ipyleaflet map.

    Parameters
    ----------
    m : ipyleaflet.Map instance
        Map instance on which the selected basemap has to be set as backdrop layer
    color : str, optional
        Color to use for the widget (default is settings.color_first)
    dark : bool, optional
        If True, the widget will have a dark background (default is settings.dark_mode)
    width : int, optional
        Width of the widget in pixels (default is 320)
    height : int, optional
        Height of the widget in pixels (default is 650)
    addBDAPbasemaps : bool, optional
        If True the treeview will contain also some BDAP layers selectable as basemaps (default is True)
    removeBasemaps : list of str, optional
        List of basemaps names to be removed from the widget (default is [])
    rootName: str, optional
        Name to use as the root node of the basemaps treeview (default is 'Basemaps')
    onchange : function, optional
        Python function to call when the user selects a different basemap. The function will receive no parameters. (default is None)
        
    Example
    -------
    Creation of a basemap selection widget::
        
        from jeodpp import inter, imap
        from ipywidgets import widgets, Layout

        from vois.vuetify import basemaps

        height = 650

        m = imap.Map(layout=Layout(height='%dpx'%height))

        b = basemaps.basemaps(m, height=height, dark=False)

        display(widgets.HBox([b.draw(),m]))

    .. figure:: figures/basemaps.png
       :scale: 100 %
       :alt: basemaps widget

       Example of a basemaps selection widget
    """
    
    def __init__(self,
                 m,
                 color=settings.color_first,
                 dark=settings.dark_mode,
                 width=320,
                 height=650,
                 addBDAPbasemaps=True,
                 removeBasemaps=[],
                 rootName='Basemaps',
                 onchange=None):
        
        self.m        = m
        self.color    = color
        self.dark     = dark
        self.width    = width
        self.height   = height
        self.onchange = onchange
        
        self.name = 'OpenStreetMap.EC'
        self.basemap_layer = basemapTileLayer(self.name)
    
        self.treecard = treeview.createTreeviewFromList(basemapList(addBDAPbasemaps=addBDAPbasemaps,removeBasemaps=removeBasemaps),
                                                        rootName=rootName, separator='.', 
                                                        expand_selection_to_parents=False,  substitutionDict={},
                                                        color=self.color, dark=self.dark, width=self.width, height=self.height, 
                                                        on_activated=self.__on_activated, selectable=False, activatable=True, 
                                                        active='OpenStreetMap.EC', opened=['Basemaps', 'OpenStreetMap'],
                                                        displayfullname=False, disabled=[])

        self.top = treeview.treeviewOperations(self.treecard)

        self.reset()
        
        
    # Set the default basemap
    def reset(self):
        """
        Set the default basemap (OpenStreetMap.EC)
        """
        self.name = 'OpenStreetMap.EC'
        self.__on_activated(self.name)
        

    # Manage activation of a node of the tree: change the basemap on the self.m map instance
    def __on_activated(self, arg):
        firstchild = self.top.getFirstChildFullname(arg)
        if firstchild:  # if the activated node has children: activate its first child
            self.top.setActive(firstchild)
        else:
            self.name = arg
            self.basemap_layer = map_setbasemap(self.m, arg)
            if not self.onchange is None:
                self.onchange()

            
    # Returns the vuetify object to display (the treeview widget)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Card containing the treeview)"""
        return self.treecard
            
        
    # current_layer property
    @property
    def current_layer(self):
        """
        Get the currently select layer (instance of ipyleaflet.leaflet.TileLayer class or ipyleaflet.leaflet.LayerGroup class)
        """
        return self.basemap_layer

    
    # value property: get and sets the current name of the basemap
    @property
    def value(self):
        """
        Get/Set the active basemap name.
        
        Returns
        --------
        name : str
            Name of the basemap

        Example
        -------
        Programmatically select one of the basemaps and print the value selected::
            
            b.value = 'Esri.WorldImagery'
            print(b.value)
        
        """
        return self.name
    
    # Set the current basemap by passing a name
    @value.setter
    def value(self, name):
        self.name = name
        self.__on_activated(self.name)
    
        


"""Utility functions for the creation of interactive maps using BDAP interactive library."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import pandas as pd
import numpy as np
import math

from ipywidgets import widgets, Layout
from ipyleaflet import SearchControl, WidgetControl, LegendControl, ScaleControl, FullScreenControl

try:
    from jeodpp import inter, imap
except:
    pass


try:
    from . import colors
    from . import geojsonUtils
except:
    import colors
    import geojsonUtils


# Custom identify
def CustomIdentifyPopup(map,p):
    from ipywidgets import widgets, HTML, CallbackDispatcher
    from ipyleaflet import Popup
    from IPython.display import display

    def handle_interaction_popup(**kwargs):
        if kwargs.get('type') == 'click':

            pos = [kwargs.get('coordinates')[0],kwargs.get('coordinates')[1]]
            message = widgets.HTML()

            sid = inter.identifyPointEx(p,pos[1],pos[0],4326,int(map.zoom))
            values = sid.split(': ')
            if len(values) == 2:
                values[1] = str(int(float(values[1])*100) / 100.0)
                sid = values[0] + ': ' + values[1]

            if len(sid) > 0:
                message.value = "<style> p.small {line-height: 1.2; }</style><p class=\"small\">" + sid.replace(",","<br />") + "</p>"
                popup = Popup(location=pos,child=message, close_button=True,auto_close=True,close_on_escape_key=True)
                map.add_layer(popup)

    map._interaction_callbacks = CallbackDispatcher()
    map.on_interaction(handle_interaction_popup)


###########################################################################################################################################################################
# Simplified way to create a vector layer displaying the countries of the world.
# Vector data is taken from inter.collections.BaseData.AdministrativeUnits.Global.VirtualEarth.Countries
# Input values to assign to countries come from a Pandas Dataframe df containing a column with iso2code and a column with values
# Returns a Map instance
###########################################################################################################################################################################
def countriesMap(df,                          # Pandas dataframe indexed on iso2_code and containing 'value' and 'label' columns
                 code_column=None,            # Name of the column containing the code of the country (None = the country code is the index of the dataframe)
                 value_column='value',        # Name of the column containing the value
                 codes_selected=[],           # codes of the countries selected
                 center=None,                 # [lat,lon] to center the map
                 zoom=None,                   # initial zoom level for the map
                 width ='99%',                # width of the map
                 height='400px',              # height of the map
                 min_width=None,              # min_width of the map
                 basemap=1,                   # Basemap to use
                 colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
                 stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
                 stroke='#232323',            # stroke color for countries border
                 stroke_selected='#00ffff',   # stroke color for border of selected country
                 stroke_width=1.0,            # border width for countries polygons
                 decimals=2,                  # Number of decimals for the legend number display
                 minallowed_value=None,       # Minimum value allowed
                 maxallowed_value=None):      # Maximum value allowed
    """
    Creation of an interactive map to display the countries of the world. An input Pandas DataFrame df is used to join a column of numeric values to the countries, using the iso2code (ISO 3166-2) as internal key attribute. Once the values are assigned to the countries, a graduated legend is calculated based on mean and standard deviation of the assigned values. A input list of colors is used to represent the countries given their assigned value.

    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it is not portable in other environments! Please refer to the module :py:mod:`leafletMap` for geospatial function not related to BDAP.

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to the countries. It has to contain at least a column with numeric values.
    code_column : str, optional
        Name of the column of the Pandas DataFrame containing the unique code of the countries in the ISO-3166-2 standard. This column is used to perform the join with the internal attribute of the countries vector dataset that contains the country code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the countries using the join on the ISO-3166-2 codes (default is 'value')
    codes_selected : list of strings, optional
        List of codes of countries to display as selected (default is [])
    center : tuple of (lat,lon), optional
        Geographical coordinates of the initial center of the interactive map visualization (default is None)
    zoom : int, optional
        Initial zoom level of the interactive map (default is None)
    width : str, optional
        Width of the map widget to create (default is '99%')
    height : str, optional
        Height of the map widget to create (default is '400px')
    min_width : str, optional
        Minimum width of the layout of the map widget (default is None)
    basemap : int, optional
        Basemap to use as background in the map visualization (default is 1). Valid values are in [1,39], see https://jeodpp.jrc.ec.europa.eu/services/processing/interhelp/3.2_map.html?highlight=basemap#inter.Map.printAvailableBasemaps for details
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to country polygons and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    stroke : str, optional
        Color to use for the border of countries (default is '#232323')
    stroke_selected : str, optional
        Color to use for the border of the selected countries (default is '#00ffff')
    stroke_width: float, optional
        Width of the border of the country polygons in pixels (default is 1.0)
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
        
    Returns
    -------
        a jeodpp.imap instance (a Map object derived from the ipyleaflet Map)

    Example
    -------
    Creation of a map displaying a random variable on 4 european countries. The numerical values assigned to each of the countries are randomly generated using numpy.random.uniform and saved into a dictionary having the country code as the key. This dict is transformed to a Pandas DataFrame with 4 rows and having 'iso2code' and 'value' as columns. The graduated legend is build using the 'inverted' Reds Plotly colorscale (low values are dark red, intermediate values are red, high values are white)::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from vois import interMap

        countries = ['DE', 'ES', 'FR', 'IT']

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        m = interMap.countriesMap(df,
                                  code_column='iso2code',
                                  height='400px',
                                  stroke_width=1.5,
                                  stroke_selected='yellow',
                                  colorlist=px.colors.sequential.Reds[::-1],
                                  codes_selected=['IT'])
        display(m)
        
    .. figure:: figures/countriesMap.png
       :scale: 100 %
       :alt: countriesMap example

       Example of an interactive map displaying 4 european countries.
            
    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)
    
    
    # Creation of the Map
    m = imap.Map(layout=Layout(width=width, height=height), basemap=basemap)
    if not min_width is None:
        m.layout.min_width = min_width
    if not center is None:
        m.center = center
    if not zoom is None:
        m.zoom = zoom

    # Add map controls
    m.add_control(FullScreenControl(position="topleft"))
    m.add_control(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))
    m.add_control(ScaleControl(position='bottomright'))
    
    # Seems not to work: to check!
    coordlabel = widgets.HTML(value='')
    widget_coordinate = WidgetControl(widget=coordlabel, position='bottomleft')
    m.add_control(widget_coordinate)
    inter.mapInteractGeneric(m, labelCoordinates=coordlabel)
    
    # Layer Countries
    v = inter.Collection(inter.collections.BaseData.AdministrativeUnits.Global.VirtualEarth.Countries)
    
    
    # Join
    countries = [str(x) for x in list(df[code_column])]
    values    = list(df[value_column])
    value_ranges = np.linspace(minvalue, maxvalue, 100)
    breaks       = list(value_ranges) # [1:]
    breaks.append(9999999999999)
    v.joinAdd('iso_a2', 'value', countries, values)   # Doesn't work in Voila!!!
    
    # Define legend
    extendedcolors = [ci.GetColor(i) for i in breaks]
    v.colorCustom(extendedcolors)
    
    v.legendSet('line', 'stroke-width', str(stroke_width))
    v.legendSet('line', 'stroke',       str(stroke))
    
    v.legendGraduated('value', 'custom', len(breaks), 2, minvalue, maxvalue, breaks)
    v.opacity(255)
    
    # Define the identify
    v.parameter('identifyField','admin value')
    v.parameter('identifyseparator', ': ')
    CustomIdentifyPopup(m, v)
    
    # Add layer to the map
    layer = m.addLayer(v.toLayer())
    
    
    # Add selected countries to the map
    if len(codes_selected) > 0:
        geoms = v.all('GEOMETRY')
        codes = v.all('iso_a2')
        for c in codes_selected:
            if c in codes:
                index = codes.index(c)
                geom = geoms[index]
                if not geom is None:
                    s = inter.VectorLayer("wkt")
                    s.geomAdd(geom)
                    s.remove('default','all')
                    s.set('line','stroke', str(stroke_selected))
                    s.set('line','stroke-width','3')
                    layersel = m.addLayer(s.toLayer(), name=c)
    
    return m
    

    
###########################################################################################################################################################################
# Load a geojson string and returns a inter.VectorLayer object
###########################################################################################################################################################################
def interGeojsonToVector(geojson):
    """
    Load a geojson string and returns a inter.VectorLayer object (see https://jeodpp.jrc.ec.europa.eu/services/processing/interhelp/3.5_vectorlayer.html)

    Parameters
    ----------
        geojson : str
            String containing data in geojson format
                
    Returns
    -------
        An instance of the inter.Vector class of the interapro library
            
    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it in not portable in other environments!
    
    """
    vector = inter.Collection(inter.collections.Vector)
    vector.fileAdd(geojson)
    return vector
    
    
###########################################################################################################################################################################
# Simplified way to create a vector layer displaying a custom geojson.
# Input values to assign to features come from a Pandas Dataframe df containing a column with unique code and a column with values
# Returns a Map instance
###########################################################################################################################################################################
def geojsonMap(df,                          # Pandas dataframe containing 'value' and 'label' columns
               geojson_path,                # Path of the geojson containing the geographic features
               geojson_attribute,           # Name of the attribute of the geojson containing the unique code
               code_column=None,            # Name of the column containing the code of the features (None = the feature code is the index of the dataframe)
               value_column='value',        # Name of the column containing the value
               codes_selected=[],           # codes of the features selected
               center=None,                 # [lat,lon] to center the map
               zoom=None,                   # initial zoom level for the map
               width ='99%',                # width of the map
               height='400px',              # height of the map
               min_width=None,              # min_width of the map
               basemap=1,                   # Basemap to use
               colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
               stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
               stroke='#232323',            # stroke color for polygons border
               stroke_selected='#00ffff',   # stroke color for border of selected polygons
               stroke_width=1.0,            # border width for polygons
               decimals=2,                  # Number of decimals for the legend number display
               minallowed_value=None,       # Minimum value allowed
               maxallowed_value=None):      # Maximum value allowed
    """
    Creation of an interactive map to display a custom geojson dataset. An input Pandas DataFrame df is used to join a column of numeric values to the geojson features, using the <geojson_attribute> as the internal key attribute. Once the values are assigned to the features, a graduated legend is calculated based on mean and standard deviation of the assigned values. A input list of colors is used to represent the featuress given their assigned value.

    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it is not portable in other environments! Please refer to the module :py:mod:`lefletMap` for geospatial function not related to BDAP.
    
    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to features. It has to contain at least a column with numeric values.
    geojson_path : str
        Path of the geojson file to load that contains the geographic features in geojson format
    geojson_attribute : str
        Name of the attribute of the geojson dataset that contains the unique codes of the features. This attribute will be use as internal key in the join operation with the df Pandas DataFrame
    code_column : str, optional
        Name of the column of the df Pandas DataFrame containing the unique code of the features. This column is used to perform the join with the internal attribute of the geojson vector dataset that contains the unique code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the features using the join on geojson unique codes (default is 'value')
    codes_selected : list of strings, optional
        List of codes of features to display as selected (default is [])
    center : tuple of (lat,lon), optional
        Geographical coordinates of the initial center of the interactive map visualization (default is None)
    zoom : int, optional
        Initial zoom level of the interactive map (default is None)
    width : str, optional
        Width of the map widget to create (default is '99%')
    height : str, optional
        Height of the map widget to create (default is '400px')
    min_width : str, optional
        Minimum width of the layout of the map widget (default is None)
    basemap : int, optional
        Basemap to use as background in the map visualization (default is 1). Valid values are in [1,39], see https://jeodpp.jrc.ec.europa.eu/services/processing/interhelp/3.2_map.html?highlight=basemap#inter.Map.printAvailableBasemaps for details
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to features and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    stroke : str, optional
        Color to use for the border of countries (default is '#232323')
    stroke_selected : str, optional
        Color to use for the border of the selected countries (default is '#00ffff')
    stroke_width: float, optional
        Width of the border of the country polygons in pixels (default is 1.0)
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
        
    Returns
    -------
        a jeodpp.imap instance (a Map object derived from the ipyleaflet Map)

    Example
    -------
    Creation of a map displaying a custom geojson. The numerical values assigned to each of the countries are randomly generated using numpy.random.uniform and saved into a dictionary having the country code as the key. This dict is transformed to a Pandas DataFrame with 4 rows and having 'iso2code' and 'value' as columns. The graduated legend is build using the 'inverted' Reds Plotly colorscale (low values are dark red, intermediate values are red, high values are white)::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from vois import interMap

        countries = ['DE', 'ES', 'FR', 'IT']

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        m = interMap.geojsonMap(df,
                                './data/ne_50m_admin_0_countries.geojson',
                                'ISO_A2_EH',   # Internal attribute used as key
                                code_column='iso2code',
                                height='400px',
                                stroke_width=1.5,
                                stroke_selected='yellow',
                                colorlist=px.colors.sequential.Reds[::-1],
                                codes_selected=['IT'])
        display(m)

    .. figure:: figures/geojsonMap.png
       :scale: 100 %
       :alt: geojsonMap example

       Example of an interactive map displaying 4 european countries from a custom geojson file.

    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)
    
    
    # Creation of the Map
    m = imap.Map(layout=Layout(width=width, height=height), basemap=basemap)
    if not min_width is None:
        m.layout.min_width = min_width
    if not center is None:
        m.center = center
    if not zoom is None:
        m.zoom = zoom

    # Add map controls
    m.add_control(FullScreenControl(position="topleft"))
    m.add_control(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))
    m.add_control(ScaleControl(position='bottomright'))
    
    # Seems not to work: to check!
    coordlabel = widgets.HTML(value='')
    widget_coordinate = WidgetControl(widget=coordlabel, position='bottomleft')
    m.add_control(widget_coordinate)
    inter.mapInteractGeneric(m, labelCoordinates=coordlabel)
    
    # Layer
    geojson = geojsonUtils.geojsonLoadFile(geojson_path)
    
    
    # Join
    countries = [str(x) for x in list(df[code_column])]
    values    = list(df[value_column])
    value_ranges = np.linspace(minvalue, maxvalue, 100)
    breaks       = list(value_ranges) # [1:]
    breaks.append(9999999999999)
    
    d = dict(zip(countries,values))
    geojsonnew = geojsonUtils.geojsonJoin(geojson,geojson_attribute, 'value', d, innerMode=True)
    v = interGeojsonToVector(geojsonnew)
    
    # Define legend
    extendedcolors = [ci.GetColor(i) for i in breaks]
    v.colorCustom(extendedcolors)
    
    v.legendSet('line', 'stroke-width', str(stroke_width))
    v.legendSet('line', 'stroke',       str(stroke))
    
    v.legendGraduated('value', 'custom', len(breaks), 2, minvalue, maxvalue, breaks)
    v.opacity(255)
    
    # Define the identify
    v.parameter('identifyField',geojson_attribute + ' value')
    v.parameter('identifyseparator', ': ')
    CustomIdentifyPopup(m, v)   
    
    # Add layer to the map
    p = v.process()
    layer = m.addLayer(v.toLayer())
    
    # Add selected countries to the map
    if len(codes_selected) > 0:
        filtered = geojsonUtils.geojsonFilter(geojsonnew, geojson_attribute, codes_selected)
        if geojsonUtils.geojsonCount(filtered) > 0:
            s = inter.Collection(inter.collections.Vector)
            s.fileAdd(filtered)
            s.remove('default','all')
            s.set('line','stroke', str(stroke_selected))
            s.set('line','stroke-width','3')
            layersel = m.addLayer(s.toLayer(), name='selected')

    return m


###########################################################################################################################################################################
# Sets a bivariate legend for vector layer v
# Returns a legend in SVG string format
###########################################################################################################################################################################
def bivariateLegend(v,
                    filters1,
                    filters2,
                    colorlist1,
                    colorlist2,
                    title='',
                    title1='',
                    title2='',
                    names1=[],
                    names2=[],
                    fontsize=14,
                    fontweight=400,
                    stroke='#000000',
                    stroke_width=0.25,
                    side=100,
                    resizewidth='',
                    resizeheight=''):
    """
    Creation of a bivariate choropleth legend for a polygon vector layer. See `Bivariate Choropleth Maps: A How-to Guide <https://www.joshuastevens.net/cartography/make-a-bivariate-choropleth-map/>`_ for the idea. The function creates a legend for vector layer v based on two attributes of the layer and returns a string containing the SVG representation of the legend (that can be displayed using display(HTML(svgstring) call)

    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it is not portable in other environments! Please refer to the module :py:mod:`lefletMap` for geospatial function not related to BDAP.
    
    Parameters
    ----------
    v : instance of inter.VectorLayer class
        Vector layer instance for which the bivariate legend has to be built
    filters1 : list of strings
        List of strings defining the conditions for the classes based on the first attribute
    filters2 : list of strings
        List of strings defining the conditions for the classes based on the second attribute
    colorlist1 : list of colors
        List of colors to use for the legend on the first attribute (see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    colorlist2 : list of colors
        List of colors to use for the legend on the second attribute (see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    title : str, optional
        Main title of the legend chart (default is '')
    title1 : str, optional
        Title for the legend on the first attribute. It will be displayed vertically in the Y axis of the SVG. Default is ''.
    title2 : str, optional
        Title for the legend on the second attribute. It will be displayed horizontally in the X axis of the SVG. Default is ''.
    names1 : list of strings, optional
        List containing one string for each of the classes of the legend on the first attribute (default is [])
    names2 : list of strings, optional
        List containing one string for each of the classes of the legend on the second attribute (default is [])
    fontsize : int, optional
        Size in pixels of the font used for texts (default is 14)
    fontweight : int, optional
        Weight of the font used for title texts (default is 400)
    stroke : str, optional
        Color to use for the border of the polygons (default is '#000000')
    stroke_width : float, optional
        Width in pixels of the stroke to use for the border of the polygons (default is 0.25)
    side : int, optional
        Side in pixels of the squares displayed in the SVG legend (default is 100)
    resizewidth : str, optional
        Width of the resizing container (default is '')
    resizeheight : str, optional
        height of the resizing container (default is '')

    Returns
    -------
        a string containing SVG text to display the bivariate legend using a call to display(HTML(...))

    Example
    -------
    Creation of a bivariate choropleth legend for the polygons of the Italian provinces. The first attribute is the short name of the province (attribute 'SIGLA'), and the second attribute is the SHAPE_AREA attribute which contains the dimension in squared meters::
    
        from ipywidgets import widgets, Layout, HTML
        from IPython.display import display
        
        from jeodpp import inter, imap
        from vois import interMap, geojsonUtils
        
        # Load data on italian provinces
        geojson = geojsonUtils.geojsonLoadFile('./data/ItalyProvinces.geojson')
        vector = interMap.interGeojsonToVector(geojson)
        vector = vector.parameter("identifyfield", "SIGLA DEN_PROV SHAPE_AREA")
        vector = vector.parameter("identifyseparator", "<br>")
        
        # Create and display a Map instance
        m = imap.Map(basemap=1, layout=Layout(height='600px'))
        display(m)
        
        # Creation of the bivariate legend
        colorlist1 = ['#f3f3f3', '#eac5dd', '#e6a3d0']
        colorlist2 = ['#f3f3f3', '#c2f1d5', '#8be2ae']

        svg = interMap.bivariateLegend(vector,
                                       ["[SIGLA] < 'FE'", "[SIGLA] >= 'FE' and [SIGLA] <= 'PU'", "[SIGLA] > 'PU'"],
                                       ["[SHAPE_AREA] < 2500000000", "[SHAPE_AREA] >= 2500000000 and [SHAPE_AREA] <= 4500000000", "[SHAPE_AREA] > 4500000000 and [SHAPE_AREA] <= 7500000000", '[SHAPE_AREA] > 7500000000'],
                                       colorlist1,
                                       colorlist2,
                                       title='Example of Bivariate Choropleth',
                                       title1="Province initials",
                                       names1=['< FE', 'in [FE,PU]', '> PU'],
                                       title2="Province area",
                                       names2=['Small', 'Medium', 'Large', 'XLarge'],
                                       fontsize=24,
                                       fontweight=500)
        
        # Display of the vector layer on the map
        p = vector.process()
        m.clear()
        m.addLayer(p.toLayer())
        m.zoomToImageExtent(p)

        inter.identifyPopup(m,p)

        # Display the bivariate choropleth legend
        display(HTML(svg))
        
    .. figure:: figures/bivariate.png
       :scale: 100 %
       :alt: Example of bivariate choropleth legend

       Example of an interactive map showing polygons colored with a bivariate choropleth legend.
        
    """
    
    n1 = len(filters1)
    n2 = len(filters2)
    v.reset()
    
    width = str(stroke_width)
    
    if n1 > 1 and n2 > 1:
        ci1 = colors.colorInterpolator(colorlist1, 0.0, n1-1)
        ci2 = colors.colorInterpolator(colorlist2, 0.0, n2-1)

        displ = 2*fontsize
        h = displ + side * n1
        w = displ + side * n2
        
        ydispl = 0
        if len(title) > 0:
            ydispl = 0.8*displ
            h += displ
            
        if len(resizewidth)  <= 0: resizewidth  = '%fpx' % w
        if len(resizeheight) <= 0: resizeheight = '%fpx' % h
            
        svg = '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 %d %d" preserveAspectRatio="xMidYMid meet" width="%s" height="calc(%s + 20px)" version="1.1">\n' % (w,h, resizewidth,resizeheight)
        
        if len(title) > 0:
            svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (displ+0.5*n2*side, 1.1*fontsize, 1.05*fontsize, fontweight, title)
            
        svg += '<text x="0" y="0" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" transform="translate(%f,%f) rotate(-90)">%s</text>' % (fontsize, fontweight, 0.8*fontsize, ydispl+0.5*n1*side, title1)
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (displ+0.5*n2*side, ydispl+2.1*fontsize+n1*side, fontsize, fontweight, title2)

        for f1 in filters1:
            row = filters1.index(f1)
            c1 = ci1.GetColor(row)
            
            if row < len(names1):
                svg += '<text x="0" y="0" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" transform="translate(%f,%f) rotate(-90)">%s</text>' % (0.7*fontsize, 400, 1.8*fontsize, ydispl+(n1-row-0.5)*side, names1[row])
                
            for f2 in filters2:
                col = filters2.index(f2)
                c2 = ci2.GetColor(col)
                cm = colors.darken(c1,c2)
                f = '(' + f1 + ') and ' + '(' + f2 + ')'
                #print(f, colors.string2rgb(c1), colors.string2rgb(c2), cm)
                v.set(f,'poly','fill',cm)
                v.set(f,'line','stroke',stroke)
                v.set(f,'line','stroke-width',width)
                
                svg += '<rect style="fill:%s;" x="%d" y="%d" width="%d" height="%d" stroke-width="0"><title>%s</title></rect>\n' % (cm, displ+col*side, ydispl+(n1-row-1)*side, side, side, f)

        for f2 in filters2:
            col = filters2.index(f2)
            if col < len(names2):
                svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (displ+(col+0.5)*side, ydispl+0.75*fontsize+n1*side, 0.7*fontsize, 400, names2[col])
                
        svg += '</svg>'
        return svg

    
###########################################################################################################################################################################
# Sets a simple trivariate legend (defined by 7 colors) using 3 attributes of a vector layer v
# Returns a legend in SVG string format
###########################################################################################################################################################################
def trivariateLegend(v,
                     filter1,
                     filter2,
                     filter3,
                     color1='#ff60ff',
                     color2='#ffff60',
                     color3='#60ffff',
                     color4='#ffffff',
                     title='',
                     title1='',
                     title2='',
                     title3='',
                     fontsize=14,
                     fontweight=300,
                     stroke='#000000',
                     stroke_width=0.25,
                     radius=100):
    """
    Creation of a trivariate choropleth legend for a polygon vector layer. See `Some Thoughts on Multivariate Maps <stamen.com/some-thoughts-on-multivariate-maps-ffe364342415/>`_ for the idea. The function creates a legend for vector layer v based on three attributes of the layer and returns a string containing the SVG representation of the legend (that can be displayed using display(HTML(svgstring) call).

    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it is not portable in other environments! Please refer to the module :py:mod:`lefletMap` for geospatial function not related to BDAP.
    
    Parameters
    ----------
    v : instance of inter.VectorLayer class
        Vector layer instance for which the trivariate legend has to be built
    filter1 : str
        Condition to filter the polygons on the first attribute
    filter2 : str
        Condition to filter the polygons on the second attribute
    filter3 : str
        Condition to filter the polygons on the third attribute
    color1 : str, optional
        Color to assign to polygons that satisfy the condition on the first attribute (default is '#ff80ff')
    color2 : str, optional
        Color to assign to polygons that satisfy the condition on the second attribute (default is '#ffff80')
    color3 : str, optional
        Color to assign to polygons that satisfy the condition on the third attribute (default is '#ff80ff')
    color4 : str, optional
        Color to assign to polygons that do not satisfy any of the three conditions (default is '#ffffff')
    title : str, optional
        Main title of the legend chart (default is '')
    title1 : str, optional
        Title for the legend on the first attribute (default is '')
    title2 : str, optional
        Title for the legend on the second attribute (default is '')
    title3 : str, optional
        Title for the legend on the third attribute (default is '')
    fontsize : int, optional
        Size in pixels of the font used for texts (default is 14)
    fontweight : int, optional
        Weight of the font used for title texts (default is 400)
    stroke : str, optional
        Color to use for the border of the polygons (default is '#000000')
    stroke_width : float, optional
        Width in pixels of the stroke to use for the border of the polygons (default is 0.25)
    radius : int, optional
        Radius in pixels of the circles displayed in the SVG legend (default is 100)

    Returns
    -------
        a string containing SVG text to display the trivariate legend using a call to display(HTML(...))

    Example
    -------
    Creation of a simple trivariate choropleth legend (with 7 colors) for a polygons layer containing crop data. The three attributes AL_PERC, PC_PERC and PG_PERC contain the percentage presence of the three specific crops inside the polygon::
    
        from ipywidgets import widgets, Layout, HTML
        from IPython.display import display
        
        from jeodpp import inter, imap
        from vois import interMap
        
        # Load data
        vector = inter.loadLocalVector("DEBY_2019_LandCover.shp")
        vector = vector.parameter("identifyfield", "LAU_NAME YEAR AL_PERC PC_PERC PG_PERC")
        vector = vector.parameter("identifyseparator", "<br>")
        
        # Create and display a Map instance
        m = imap.Map(basemap=60, layout=Layout(height='600px'))
        display(m)
        
        # Creation of the bivariate legend
        svg = interMap.trivariateLegend(vector,
                                        "[AL_PERC] > 60",
                                        "[PC_PERC] > 10",
                                        "[PG_PERC] > 20",
                                        '#ff60ff',
                                        '#ffff60',
                                        '#60ffff',
                                        '#ffffff55',
                                        title='Example of Trivariate Choropleth',
                                        title1="Arable Land",
                                        title2="Perm. Crop",
                                        title3="Permanent Grassland",
                                        fontsize=12,
                                        fontweight=500,
                                        radius=70)
        
        # Display of the vector layer on the map
        p = vector.process()
        m.clear()
        m.addLayer(p.toLayer())
        m.zoomToImageExtent(p)

        inter.identifyPopup(m,p)

        # Display the trivariate choropleth legend
        display(HTML(svg))
        
    .. figure:: figures/trivariate.png
       :scale: 100 %
       :alt: Example of trivariate choropleth legend

       Example of an interactive map showing polygons colored with a trivariate choropleth legend.
        
    """
    
    width = str(stroke_width)

    # Set the legend
    v.reset()
    v.set('all','poly','fill',color4)
    
    v.set(filter2,'poly','fill',color3)
    v.set(filter2,'line','stroke',stroke)
    v.set(filter2,'line','stroke-width',width)

    v.set(filter2,'poly','fill',color2)
    v.set(filter2,'line','stroke',stroke)
    v.set(filter2,'line','stroke-width',width)
    
    v.set(filter1,'poly','fill',color1)
    v.set(filter1,'line','stroke',stroke)
    v.set(filter1,'line','stroke-width',width)

    f = '(' + filter1 + ') and ' + '(' + filter2 + ')'
    c = colors.darken(color1,color2)
    v.set(f,'poly','fill',c)
    v.set(f,'line','stroke',stroke)
    v.set(f,'line','stroke-width',width)
    
    f = '(' + filter1 + ') and ' + '(' + filter3 + ')'
    c = colors.darken(color1,color3)
    v.set(f,'poly','fill',c)
    v.set(f,'line','stroke',stroke)
    v.set(f,'line','stroke-width',width)
    
    f = '(' + filter2 + ') and ' + '(' + filter3 + ')'
    c = colors.darken(color2,color3)
    v.set(f,'poly','fill',c)
    v.set(f,'line','stroke',stroke)
    v.set(f,'line','stroke-width',width)    
    
    f = '(' + filter1 + ') and ' + '(' + filter2 + ') and ' + '(' + filter3 + ')'
    c = colors.darken(color1,colors.darken(color2,color3))
    v.set(f,'poly','fill',c)
    v.set(f,'line','stroke',stroke)
    v.set(f,'line','stroke-width',width)
    
    # Create the SVG
    displ = 1.5*fontsize
    h = displ + 2.6*radius + displ
    w = displ + 2.6*radius + displ

    ydispl = 0
    if len(title) > 0:
        ydispl = 0.8*displ
        h += displ

    svg = '<svg xmlns="http://www.w3.org/2000/svg" width="%d" height="%d" version="1.1">\n' % (w, h)

    if len(title) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (0.5*w, 1.1*fontsize, 1.05*fontsize, fontweight, title)

    svg += '<circle cx="%f" cy="%f" r="%f" fill="%s"></circle>'                                 % (0.333*w, 0.666*h, radius, color1)
    svg += '<circle cx="%f" cy="%f" r="%f" fill="%s" style="mix-blend-mode: darken;"></circle>' % (0.666*w, 0.666*h, radius, color2)
    svg += '<circle cx="%f" cy="%f" r="%f" fill="%s" style="mix-blend-mode: darken;"></circle>' % (0.500*w, 0.390*h, radius, color3)
    
    svg += '<text x="%f" y="%f" text-anchor="end"    font-size="%f" fill="black" font-weight="%d" >%s</text>' % (0.333*w, 0.666*h+fontsize,     fontsize, fontweight, title1)
    svg += '<text x="%f" y="%f" text-anchor="start"  font-size="%f" fill="black" font-weight="%d" >%s</text>' % (0.666*w, 0.666*h+fontsize,     fontsize, fontweight, title2)
    svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (0.500*w, 0.390*h-1.5*fontsize, fontsize, fontweight, title3)


    svg += '</svg>'
    return svg
    
    
###########################################################################################################################################################################
# Sets a complex trivariate legend for vector layer v based on three numerical attributes
# Returns a legend in SVG string format
###########################################################################################################################################################################
def trivariateLegendEx(v,
                       attribute1,
                       attribute2,
                       attribute3,
                       n=3,
                       min1=0.0,
                       max1=100.0,
                       min2=0.0,
                       max2=100.0,
                       min3=0.0,
                       max3=100.0,
                       color1='#ff80f7',
                       color2='#00d1d0',
                       color3='#cfb000',
                       color4='#ffffff',
                       title='',
                       title1='',
                       title2='',
                       title3='',
                       fontsize=14,
                       fontweight=400,
                       stroke='#000000',
                       stroke_width=0.25,
                       side=200,
                       resizewidth='',
                       resizeheight='',
                       digits=2,
                       maxticks=0,
                       showarrows=True):
    """
    Creation of a trivariate choropleth legend for a polygon vector layer. See `Choropleth maps with tricolore <https://cran.r-project.org/web/packages/tricolore/vignettes/choropleth_maps_with_tricolore.html>`_ for the idea. The function creates a legend for vector layer v based on three attributes of the layer and returns a string containing the SVG representation of the legend in the form of a triangle (that can be displayed using display(HTML(svgstring) call)

    Note
    ----
    This function is built on top of the BDAP interapro library to display dynamic geospatial dataset. For this reason it is not portable in other environments! Please refer to the module :py:mod:`lefletMap` for geospatial function not related to BDAP.
    
    Parameters
    ----------
    v : instance of inter.VectorLayer class
        Vector layer instance for which the trivariate legend has to be built
    attribute1 : str
        Name of the first numerical attribute
    attribute2 : str
        Name of the second numerical attribute
    attribute3 : str
        Name of the third numerical attribute
    n : int, optional
        Number of intervals for each of the three numerical attributes (default is 3). Accepptable values are those in the range [2,10]
    min1 : float, optional
        Minimum value for the first attribute (default is 0.0)
    max1 : float, optional
        Maximum value for the first attribute (default is 100.0)
    min2 : float, optional
        Minimum value for the second attribute (default is 0.0)
    max2 : float, optional
        Maximum value for the second attribute (default is 100.0)
    min3 : float, optional
        Minimum value for the third attribute (default is 0.0)
    max3 : float, optional
        Maximum value for the third attribute (default is 100.0)
    color1 : str, optional
        Color to assign to polygons that have the maximum value on the first attribute (default is '#ff80f7')
    color2 : str, optional
        Color to assign to polygons that have the maximum value on the second attribute (default is '#00d1d0')
    color3 : str, optional
        Color to assign to polygons that have the maximum value on the third attribute (default is '#cfb000')
    color4 : str, optional
        Color to assign to polygons that have all the three values of the attributes smaller than the corresponding minimal value (default is '#ffffff'). For this color, the transparency can be set, for instance using '#ffffff88' for partial transparency or '#ffffffff' for full transparency.
    title : str, optional
        Main title of the legend chart (default is '')
    title1 : str, optional
        Title for the legend on the first attribute. It will be displayed on the bottom side of the triangle SVG. Default is ''.
    title2 : str, optional
        Title for the legend on the second attribute. It will be displayed on the right side of the triangle SVG. Default is ''.
    title3 : str, optional
        Title for the legend on the third attribute. It will be displayed on the left side of the triangle SVG. Default is ''.
    fontsize : int, optional
        Size in pixels of the font used for texts (default is 14)
    fontweight : int, optional
        Weight of the font used for title texts (default is 400)
    stroke : str, optional
        Color to use for the border of the polygons (default is '#000000')
    stroke_width : float, optional
        Width in pixels of the stroke to use for the border of the polygons (default is 0.25)
    side : int, optional
        Dimension in pixels of one side of the triangle displayed in the SVG legend (default is 200)
    resizewidth : str, optional
        Width of the resizing container (default is '')
    resizeheight : str, optional
        height of the resizing container (default is '')
    digits : int, optional
        Number of decimal digits to use for displaying numerical values on the axis of the SVG chart (default is 2)
    maxticks: int, optional
        Maximum number of tick marks to display on each of the triangle sides. If 0 or less, the ticks for all the intervals are shown. Default is 0
    showarrows : bool, optional
        If True displays small arrows to help identify the three axes (default is True)

    Returns
    -------
        a string containing SVG text to display the trivariate legend using a call to display(HTML(...))

    Example
    -------
    Creation of a complex trivariate choropleth legend for a polygons layer containing crop data. The three attributes AL_PERC, PC_PERC and PG_PERC contain the percentage presence of the three specific crops inside the polygon::
    
        from ipywidgets import widgets, Layout, HTML
        from IPython.display import display
        
        from jeodpp import inter, imap
        from vois import interMap
        
        # Load data
        vector = inter.loadLocalVector("DEBY_2019_LandCover.shp")
        vector = vector.parameter("identifyfield", "LAU_NAME YEAR AL_PERC PC_PERC PG_PERC")
        vector = vector.parameter("identifyseparator", "<br>")
        
        # Create and display a Map instance
        m = imap.Map(basemap=60, layout=Layout(height='600px'))
        display(m)
        
        svg = interMap.trivariateLegendEx(vector,
                                          "AL_PERC",
                                          "PC_PERC",
                                          "PG_PERC",
                                          6,
                                          0.0,
                                          100.0,
                                          0.0,
                                          100.0,
                                          0.0,
                                          100.0,
                                          color1='#ff80f7',
                                          color2='#00d1d0',
                                          color3='#cfb000',
                                          color4='#ffffff00',
                                          title='Complex Trivariate Choropleth',
                                          title1="Arable Land",
                                          title2="Perm. Crop",
                                          title3="Perm. Grassl.",
                                          fontsize=18,
                                          fontweight=500,
                                          side=400,
                                          digits=0,
                                          maxticks=5,
                                          showarrows=True)


        # Display of the vector layer on the map
        p = vector.process()
        m.clear()
        m.addLayer(p.toLayer())
        m.zoomToImageExtent(p)

        inter.identifyPopup(m,p)

        # Display the trivariate choropleth legend
        display(HTML(svg))
        
        
    .. figure:: figures/trivariateex.png
       :scale: 100 %
       :alt: Example of complex trivariate choropleth legend

       Example of an interactive map showing polygons colored with a complex trivariate choropleth legend.
       
    """

    if n < 2:  n = 2
    if n > 10: n = 10
    
    # Convert colors to (r,g,b) tuples
    rgb1 = colors.string2rgb(color1)
    rgb2 = colors.string2rgb(color2)
    rgb3 = colors.string2rgb(color3)
    
    # Maximum distance among small triangles centers
    maxdistance = side * (n - 1) / float(n)
    
    # Steps on the three attributes ranges
    step1 = (max1 - min1) / n
    step2 = (max2 - min2) / n
    step3 = (max3 - min3) / n
    
    
    # Interpolate a color from 3 percentage distance values (d1,d2,d3 in [0.0,1.0])
    def colorInterpolate(d1, d2, d3):
        
        # Weights of the three colors
        w1 = 1.0 - d1
        w2 = 1.0 - d2
        w3 = 1.0 - d3
        
        wtot = w1 + w2 + w3
        if wtot <= 0.000000001:
            return color4
        
        w1 /= wtot
        w2 /= wtot
        w3 /= wtot
        
        r = int(rgb1[0]*w1 + rgb2[0]*w2 + rgb3[0]*w3)
        g = int(rgb1[1]*w1 + rgb2[1]*w2 + rgb3[1]*w3)
        b = int(rgb1[2]*w1 + rgb2[2]*w2 + rgb3[2]*w3)
        return colors.rgb2hex((r,g,b))

                    
    # Returns the list of n points at equal distance on the segment (x1,y1)->(x2,y2)
    def pointOnSegment(p1,p2, n):
        points = [p1]
        d = 1.0/float(n)
        for i in range(1,n):
            k = d*i
            x = k*p2[0] + (1.0-k)*p1[0]
            y = k*p2[1] + (1.0-k)*p1[1]
            points.append((x,y))
        points.append(p2)
        return points
        
    
    # Returns the center point of a triangle
    def center(p1, p2, p3):
        return ((p1[0]+p2[0]+p3[0])/3.0, (p1[1]+p2[1]+p3[1])/3.0)

    # Returns the distance between two points
    def distance(p1,p2):
        return math.hypot(p2[0] - p1[0], p2[1] - p1[1])
    
    
    # Returns the SVG string to display a filled triangle
    def triangle(p1, p2, p3):
        c = center(p1,p2,p3)
        d1 = distance(c,P1)
        d2 = distance(c,P2)
        d3 = distance(c,P3)
        if d1 <= 0.001*side:
            d1 = 0.0
            color = color1
        else:
            if d2 <= 0.001*side:
                d2 = 0.0
                color = color2
            else:
                if d3 <= 0.001*side:
                    d3 = 0.0
                    color = color3
                else:
                    d1 /= maxdistance
                    d2 /= maxdistance
                    d3 /= maxdistance
                    color = colorInterpolate(d1,d2,d3)
        
        d = side*0.0025
        x1 = p1[0]
        y1 = p1[1]
        x2 = p2[0]
        y2 = p2[1]
        x3 = p3[0]
        y3 = p3[1]
        
        # Slightly enlarge the triangles so they overlap
        if y1 > y3:
            x1 -= d
            x2 += d
            y1 += d
            y2 += d
            y3 -= d
        else:
            x1 -= d
            x2 += d
            y1 -= d
            y2 -= d
            y3 += d
            
        return '<path d="M%f %f L%f %f L%f %f z" stroke-width="0" fill="%s"/>' % (x1,y1, x2,y2, x3,y3, color)

    
    # Maps a value x from [xmin,ymin] to [x1,x2]
    def linearMap(x,xmin,xmax, x1,x2):
        if   x <= xmin: return x1
        elif x >= xmax: return x2
        else:
            p = (x - xmin)/(xmax - xmin)
            return (1.0 - p)*x1 + p*x2
        
        
    # From a step index returns the interval on one of the attributes plus the distance from the corresponding full color
    def stepToValues(index, minvalue, maxvalue, step, attribname):
        if index == 0:
            return "[" + attribname + "] < "  + str(minvalue+step), maxdistance
        elif index >= n-1:
            return "[" + attribname + "] >= " + str(maxvalue-step), 0.0
        else:
            return "[" + attribname + "] >= " + str(minvalue+index*step) + " and [" + attribname + "] < " + str(minvalue+(index+1)*step), (n-index-1)*maxdistance/float(n)

        
    # Legend creation for the vector layer
    width = str(stroke_width)
    v.reset()
    v.set('all','poly','fill',color4)
    v.set('all','line','stroke',stroke)
    v.set('all','line','stroke-width',width)

    for i1 in range(n):
        for i2 in range(n):
            for i3 in range(n):
                filter1,d1 = stepToValues(i1, min1, max1, step1, attribute1)
                filter2,d2 = stepToValues(i2, min2, max2, step2, attribute2)
                filter3,d3 = stepToValues(i3, min3, max3, step3, attribute3)
                filterall = filter1 + " and " + filter2 + " and " + filter3
                
                if d1 <= 0.001*side:
                    d1 = 0.0
                    color = color1
                else:
                    if d2 <= 0.001*side:
                        d2 = 0.0
                        color = color2
                    else:
                        if d3 <= 0.001*side:
                            d3 = 0.0
                            color = color3
                        else:
                            d1 /= maxdistance
                            d2 /= maxdistance
                            d3 /= maxdistance
                            color = colorInterpolate(d1,d2,d3)
                v.set(filterall,'poly','fill',color)
                v.set(filterall,'line','stroke',stroke)
                v.set(filterall,'line','stroke-width',width)
    
    displ = 4*fontsize
    tside = side
    th    = side * 0.5 * math.sqrt(3.0)
    w = displ + tside + displ
    h = 0.5*displ + th + 0.5*displ
        
    if len(resizewidth)  <= 0: resizewidth  = '%fpx' % w
    if len(resizeheight) <= 0: resizeheight = '%fpx' % h
        
    ydispl = 0
    if len(title) > 0:
        ydispl = 0.8*displ
        h += displ

    x1 = displ
    x2 = x1 + tside
    x3 = 0.5*(x1+x2)
    y1 = y2 = h - displ
    y3 = y1 - th
            
    svg = '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 %d %d" preserveAspectRatio="xMidYMid meet" width="%s" height="calc(%s + 20px)" version="1.1">\n' % (w,h, resizewidth,resizeheight)

    if len(title) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (x3, 1.25*fontsize, 1.05*fontsize, fontweight, title)

    # Height of the small triangles
    sh = th/float(n)
        
    # Returns the third point of a triangle having p1 and p2 as the base points
    def third(p1,p2, h):
        return (0.5*(p1[0]+p2[0]), p1[1]+h)
        
        
    # Center points of the three triangles on the three corners
    points = pointOnSegment((x1,y1),(x2,y2),n)
    pa = points[-2]
    pb = points[-1]
    pc = third(pa,pb,-sh)
    P1 = center(pa,pb,pc)
        
    points = pointOnSegment((x2,y2),(x3,y3),n)
    pa = (points[-2][0]-side/n,points[-2][1])
    pb = points[-2]
    pc = points[-1]
    P2 = center(pa,pb,pc)
        
    points = pointOnSegment((x1,y1),(x2,y2),n)
    pa = points[0]
    pb = points[1]
    pc = third(pa,pb,-sh)
    P3 = center(pa,pb,pc)

    # Creation of the small triangles
    p1 = (x1,y1)
    p2 = (x2,y2)
    for level in range(n):
        points = pointOnSegment(p1,p2,n-level)
        for i in range(n-level):
            p1 = points[i]
            p2 = points[i+1]
            svg += triangle(p1,p2,third(p1,p2,-sh))
            if i < n-level-1:
                p3 = points[i+2]
                svg += triangle(third(p1,p2,-sh),third(p2,p3,-sh),p2)

        p1 = third(points[0], points[1], -sh)
        p2 = third(points[-2],points[-1],-sh)
                
        
    # External border pf the triangle
    swidth = side/140.0
    d = side*0.005
    svg += '<path d="M%f %f L%f %f L%f %f z" stroke="black" stroke-width="%f" fill="none"/>' % (x1-d,y1+d, x2+d,y2+d, x3,y3-d, swidth)
        
        
    # Text for the three attributes
    if len(title1) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (x2, y2+2.5*fontsize, fontsize, fontweight, title1)
    if showarrows:
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (x2-2*fontsize, y2+1.45*fontsize, x2, y2+1.45*fontsize)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (x2, y2+1.45*fontsize, x2-0.2*fontsize, y2+1.25*fontsize)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (x2, y2+1.45*fontsize, x2-0.2*fontsize, y2+1.65*fontsize)
        
    if len(title2) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (x3, y3-1.1*fontsize, fontsize, fontweight, title2)
    if showarrows:
        points = pointOnSegment((x3,y3),(x2,y2),10)
        ax1 = points[0][0]+3.2*fontsize
        ax2 = points[1][0]+3.2*fontsize
        ay1 = points[0][1]+fontsize
        ay2 = points[1][1]+fontsize
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax2,ay2)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax1+0.3*fontsize,ay1)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax1-0.15*fontsize,ay1+0.3*fontsize)
        
    if len(title3) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="%d" >%s</text>' % (x1, y1+2.5*fontsize, fontsize, fontweight, title3)
    if showarrows:
        points = pointOnSegment((x1,y1),(x3,y3),10)
        ax1 = points[0][0]-2.2*fontsize
        ax2 = points[1][0]-2.2*fontsize
        ay1 = points[0][1]-fontsize
        ay2 = points[1][1]-fontsize
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax2,ay2)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax1+0.36*fontsize,ay1-0.05*fontsize)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke="black" stroke-width="1.0" fill="black"/>' % (ax1,ay1,ax1-0.15*fontsize,ay1-0.35*fontsize)
        
    # Texts on the three axes
    f = "{:.%df}" % digits
        
    nticks = n
    if maxticks > 0 and maxticks < nticks: nticks = maxticks
    points = pointOnSegment((x1,y1),(x2,y2),nticks)
    for index, p in enumerate(points):
        v = linearMap(index, 0, nticks, min1, max1)
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" fill="black" font-weight="300" >%s</text>' % (p[0], p[1]+1.1*fontsize, 0.9*fontsize, f.format(v))
        
    points = pointOnSegment((x2,y2),(x3,y3),nticks)
    for index, p in enumerate(points):
        v = linearMap(index, 0, nticks, min2, max2)
        svg += '<text x="%f" y="%f" text-anchor="start" font-size="%f" fill="black" font-weight="300" >%s</text>' % (p[0]+0.3*fontsize, p[1], 0.9*fontsize, f.format(v))
            
    points = pointOnSegment((x3,y3),(x1,y1),nticks)
    for index, p in enumerate(points):
        v = linearMap(index, 0, nticks, min3, max3)
        svg += '<text x="%f" y="%f" text-anchor="end" font-size="%f" fill="black" font-weight="300" >%s</text>' % (p[0]-0.3*fontsize, p[1], 0.9*fontsize, f.format(v))
            
        
    svg += '</svg>'
    return svg
    


"""Input widget to select a date"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from datetime import datetime
from vois.vuetify.utils.util import *
from typing import Callable, Optional

import warnings


class DatePicker(v.Menu):
    """
    Input widget to select a date.

    Parameters
    ----------
    date : str, optional
        Initial date selected on the input widget expressed in format 'YYYY-MM-DD' (default is None, corresponding to the today date)
    label : str, optional
        Label to be displayed inside the widget (default is '')
    dark : bool, optional
        If True, the popup date selection will have a dark background (default is settings.dark_mode)
    width : int, optional
        Width of the widget in pixels (default is 88)
    color : str, optional
        Color to use for the widget and the header of the popup window (default is settings.color_first)
    show_week : bool, optional
        If True the popup window will also show number of the week (default is False)
    on_change : function, optional
        Python function to call when the user selects a date. The function will receive no parameters. (default is None)
    offset_x : bool, optional
        If True the popup window will be opened on the right of the input field (default is False)
    offset_y : bool, optional
        If True the popup window will be opened on the bottom of the input field (default is True)
    disabled : bool, optional
        True if the selection of the date is disabled, False otherwise (default is False)
    min_date : str, optional
        Minimum selectable date in format 'YYYY-MM-DD' (default is None)
    max_date : str, optional
        Maximum selectable date in format 'YYYY-MM-DD' (default is None)

    Example
    -------
    Creation of a date picker widget::
        
        from vois.vuetify import DatePicker
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange():
            with output:
                print('Changed to', d.date)

        d = DatePicker(
            date=None,
            label='Start date',
            offset_x=True,
            offset_y=False,
            onchange=onchange)

        display(d)
        display(output)

    .. figure:: figures/datePicker.png
       :scale: 100 %
       :alt: datePicker widget

       Example of a datePicker
    """
    deprecation_alias = dict(mindate='min_date', maxdate='max_date', onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 date: Optional[str] = None,
                 label: str = '',
                 dark: bool = None,
                 width: int = 88,
                 color: str = None,
                 show_week: bool = False,
                 on_change: Optional[Callable[[], None]] = None,
                 offset_x: bool = False,
                 offset_y: bool = True,
                 disabled: bool = False,
                 min_date: Optional[str] = None,
                 max_date: Optional[str] = None):

        from vois.vuetify import settings

        if date is None:
            self._date = datetime.today().strftime('%Y-%m-%d')
        else:
            self._date = date

        self.label = label
        self.dark = dark if dark is not None else settings.dark_mode
        self.width = width
        self.color = color if color is not None else settings.color_first
        self.show_week = show_week
        self.on_change = on_change
        self.offset_x = offset_x
        self.offset_y = offset_y
        self._disabled = disabled
        self.min_date = min_date
        self.max_date = max_date

        margins = "mb-n5"
        if len(self.label) == 0: margins = "mb-n5 mt-n2"

        if self._disabled:
            von = ''
        else:
            von = 'menuData.on'
        self.tf = v.TextField(v_on=von, style_="max-width: %dpx; overflow: hidden;" % self.width, v_model=self._date,
                              type_="date", color=self.color, readonly=True, label=self.label, class_=margins)

        self.ctf = v.Card(flat=True, style_="max-width: %dpx; overflow: hidden;" % self.width, class_="pa-0 ma-0",
                          children=[self.tf])

        self.p = v.DatePicker(v_model=self._date, flat=True, elevation=0, color=self.color, header_color=self.color,
                              style_="max-width: 290px;", show_week=self.show_week, dark=self.dark)

        if self.min_date:
            self.p.min = self.min_date
        if self.max_date:
            self.p.max = self.max_date

        self.p.on_event('input', self.__internal_onchange)

        super().__init__(offset_x=self.offset_x, offset_y=self.offset_y, open_on_hover=False, dense=True,
                         close_on_click=True, close_on_content_click=False,
                         v_slots=[{'name': 'activator', 'variable': 'menuData', 'children': [self.ctf]}])

        self.children = [self.p]

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Manage 'input' event
    def __internal_onchange(self, widget, event, data):
        self._date = data
        self.tf.v_model = data
        if self.on_change:
            self.on_change()

    # Returns the vuetify object to display (the v.Menu)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self

    # date property
    @property
    def date(self):
        """
        Get/Set the selected date.
        
        Returns
        --------
        d : str
            date currently selected in the format 'YYYY-MM-DD'

        Example
        -------
        Programmatically change the date::
            
            picker.date = '2022-07-15'
            print(picker.date)
        
        """
        return self._date

    @date.setter
    def date(self, d):
        if isinstance(d, str) and len(d) == 10 and d[4] == '-' and d[7] == '-':
            self._date = d
            self.tf.v_model = self._date
            self.p.v_model = self._date
            if self.on_change:
                self.on_change()
        else:
            self._date = None
            self.tf.v_model = self._date
            self.p.v_model = self._date

    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state of the widget.
        
        Returns
        --------
        flag : bool
            True if the widget is disabled, False otherwise

        Example
        -------
        Programmatically change the date::
            
            picker.disabled = True
            print(picker.disabled)
        
        """
        return self._disabled

    @disabled.setter
    def disabled(self, flag):
        self._disabled = bool(flag)
        if self._disabled:
            von = ''
        else:
            von = 'menuData.on'
        self.tf.v_on = von



"""BDAP (JRC Big Data Analytics Platform) layer creation for raster display and identify on a ipyleaflet.Map without dependencies on interapro library. Derived class of serverlayer.serverlayer."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Widgets import
import ipyleaflet
from ipywidgets import widgets, HTML, CallbackDispatcher
from ipyleaflet import Popup

# Python imports
from io import StringIO, BytesIO
import sys
import json
import requests
import numpy as np

# Local imports
import serverlayer


#####################################################################################################################################################
# Class layer
#####################################################################################################################################################
class layer(serverlayer.serverlayer):
    
    # Initialize
    def __init__(self,
                 filepath='',
                 band=1,
                 epsg=4326,
                 nodata=999999.0,
                 identify_integer=False):
        
        super().__init__(filepath=filepath, band=band, epsg=epsg, nodata=nodata, identify_integer=identify_integer)
        
        # Dict to display strings instead of integers on identify
        self.identify_dict = None

    
    # Returns an instance of ipyleaflet.TileLayer
    def tileLayer(self, max_zoom=22):
        url = self.tileUrl()
        if not url is None:
            return ipyleaflet.TileLayer(url=url, max_zoom=max_zoom, max_native_zoom=max_zoom)
    

    # Execute an identify operation server-side using HTTP call. Returns a scalar float/int/string or a list of scalars
    def identifyPoint(self, lon, lat, epsg=4326, zoomlevel=10):
        response = requests.get('https://jeodpp.jrc.ec.europa.eu/jiplib-view/?IDENTIFYEX=1&vl=%s&x=%f&y=%f&epsg=%d&zoom=%d' % (self.toLayer(), lon, lat, epsg, int(zoomlevel)))
        bio = BytesIO(response.content)
        s = bio.read()
        svalue = s.decode("utf-8").replace('Values = ','').replace('Value = ','')
        if len(svalue) > 0:
            if ',' in svalue:
                vvv = svalue.split(',')
                if len(vvv) <= len(self.ratios):
                    if self.identify_integer:
                        return [int(round(serverlayer.unscaled(float(x),r,o))) for x,r,o in zip(vvv,self.ratios,self.offsets)]
                    else:
                        return [serverlayer.unscaled(float(x),r,o) for x,r,o in zip(vvv,self.ratios,self.offsets)]
                else:
                    if self.identify_integer:
                        return [int(x) for x in vvv]
                    else:
                        return [float(x) for x in vvv]
            else:
                if len(self.ratios) >= 1 and len(self.offsets) >= 1:
                    if self.identify_integer:
                        return serverlayer.unscaled(float(svalue),self.ratios[0],self.offsets[0])
                    else:
                        return int(round(serverlayer.unscaled(float(svalue),self.ratios[0],self.offsets[0])))
                else:
                    if self.identify_integer:
                        ivalue = int(svalue)
                        if not self.identify_dict is None and ivalue in self.identify_dict:
                            return self.identify_dict[ivalue]
                        return ivalue
                    else:
                        return float(svalue)
    

    # Set-up the identify action on a map for the layer
    def identifySetup(self, m, onclick=None, identify_dict=None):

        def handle_interaction_popup(**kwargs):
            if kwargs.get('type') == 'click':
                lat = kwargs.get('coordinates')[0]
                lon = kwargs.get('coordinates')[1]
                res = self.identifyPoint(lon, lat, zoomlevel=int(m.zoom))  # res can be: None, a floating scalar, a list of float values
                if not res is None:
                    if onclick is None:
                        message = widgets.HTML()
                        message.value = "<style> p.small {line-height: 1.2; }</style><p class=\"small\">" + str(res) + "</p>"
                        popup = Popup(location=[lat,lon], child=message, close_button=True, auto_close=True, close_on_escape_key=True)
                        m.add_layer(popup)
                    else:
                        onclick(res, lon, lat)

        self.identify_dict = identify_dict
        m._interaction_callbacks = CallbackDispatcher()
        m.on_interaction(handle_interaction_popup)        


"""Popup window opened at hover on a button."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

from random import randint

try:
    from . import settings
    from . import fontsettings
except:
    import settings
    import fontsettings


#####################################################################################################################################################
# Popup control
#####################################################################################################################################################
class popup:
    """
    Popup window opened at hover on a button
        
    Parameters
    ----------
    widget : any widget
        Widget to be displayed inside the popup menu
    buttontext : str
        Text to display inside the button
    buttonwidth : int, optional
        Width of the button in pixels (default is 140)
    buttonheight : int, optional
        Height of the button in pixels (default is 40)
    icon : str, optional
        Name of the icon to be displayed inside the button (default is None)
    icon_small : bool, optional
        If True the icon will be small (default is True)
    icon_large : bool, optional
        If True the icon will be small (default is False)
    margins : int, optional
        Dimension of the margins on all directions (default is 0)
    margintop : int, optional
        Dimension of the margin on top of the label (default is None)
    color : str, optional
        Color used for the button (default is the color_first defined in the settings.py module)
    rounded : bool, optional
        If True the button will be rounded (default is the button_rounded defined in the settings.py module)
    outlined : bool, optional
        If True the button will be outlined (default is True)
    text : bool, optional
        If True the button will display only the text and/or the icon, with no background (default is False)
    popupwidth : int, optional
        Width of the popup window in pixels (default is 160). The popup cannot have a width smaller than the width of the button
    popupheight : int, optional
        Height of the popup window in pixels (default is 250)
    open_on_hover : bool, optional
        If True the popup opens/closes when hovering the button, otherwise it opens/closes on click on the button (default is True)
    close_on_click : bool, optional
        Designates if popup should close on outside click (default is True)
    close_on_content_click : bool, optional
        Designates if popup should close when its content is clicked  (default is True)
    title: str, optional
        Text to add at the top of the popup (default is '')
    show_close_button : bool, optional
        If True a close icon button is displayed on the top-right side of the popup to ease the closing of the popup (default is False). It can be useful mainly when close_on_click is set to False.
    on_input : function, optional
        Python function to call when the popup opens or closes by click or by hover (default is None). When the popup opens the function will be called with True as parameter, while when the popup menu closes it will be called with False as parameter

    Note
    ----
    All the icons from https://materialdesignicons.com/ site can be used, just by prepending 'mdi-' to their name.
    
    All the free icons from https://fontawesome.com/ site can be used, just by prepending 'fa-' to their name.
    
    
    Example
    -------
    Creation and display of a popup window displaying a tree and opened by hovering on a button::
        
        from vois.vuetify import popup, treeview
        from IPython.display import display

        sectors = ['S%d'%x for x in range(10)]
        treecard = treeview.createTreeviewFromList(sectors,
                                                   rootName='All',
                                                   height='270px')

        p = popup.popup(treecard, 'Sectors', popupheight=270)

        display(p.draw())

    .. figure:: figures/popup.png
       :scale: 100 %
       :alt: popup widget example

       Example of a button that, on hover, opens a popup window containing a treeview
   """

    # Initialization
    def __init__(self,
                 widget,
                 buttontext,
                 buttonwidth=140,
                 buttonheight=40,
                 icon=None,
                 icon_small=True,
                 icon_large=False,
                 margins=0,
                 margintop=None,
                 color=settings.color_first,
                 rounded=settings.button_rounded,
                 outlined=True,
                 text=False,
                 popupwidth=160,
                 popupheight=250,
                 open_on_hover=True,
                 close_on_click=True,
                 close_on_content_click=True,
                 title='',
                 show_close_button=False,
                 on_input=None):
        
        self.buttontext = buttontext
        self.icon       = icon
        self.icon_small = icon_small
        self.icon_large = icon_large
        self.on_input   = on_input

        # The popup cannot have a width smaller than the width of the button
        if popupwidth < buttonwidth: popupwidth = buttonwidth
            
        # Add a close button on top
        children = [widget]
        if show_close_button or len(title)> 0:
            
            if show_close_button:
                def onclick(*args):
                    self.menu.v_model = False
                    
                bclose = v.Btn(icon=True, small=True, children=[v.Icon(small=True, children=['mdi-close'])])
                bclose.on_event('click', onclick)
            else:
                bclose = v.Html(tag='div',children=[''])
            
            htitle = v.Html(tag='div',children=[title], class_='pa-0 ma-1', style_='width: %dpx; height: 1px;'%(popupwidth-38))
                                
            r = v.Row(no_gutters=True, justify="start", children=[htitle,bclose], class_='pa-0 ma-0')
            children = [r, widget]
            popupheight += 30
            
        card = v.Card(width='%dpx'%popupwidth, height='%dpx'%popupheight, elevation=1, children=children, style_='overflow: auto;')
        
        children, flagicon = self.__setup_btn_children()
        
        if not margintop is None:
            class_ = "pa-0 ma-%s mt-%s" % (str(margins), str(margintop))
        else:
            class_ = "pa-0 ma-%s" % (str(margins))
            
        variable = 'menuData%d' % randint(1,999999)
        
        self.btn = v.Btn(v_on='%s.on'%variable, color=color, fab=False, dark=True, depressed=True, icon=flagicon, text=text, class_=class_,
                         disabled=False, width=buttonwidth, max_width=buttonwidth, height=buttonheight, rounded=rounded, outlined=outlined,
                         style_='font-family: %s; font-weight: %d; text-transform: none' % (fontsettings.font_name, 450),
                         children=children)
        
        self.menu = v.Menu(v_model=False, offset_y=True, open_on_hover=open_on_hover, dense=True, internal_activator=True,
                           close_on_click=close_on_click, close_on_content_click=close_on_content_click,
                           v_slots=[{'name': 'activator', 'variable': variable, 'children': self.btn }],
                           children=[card], style_='overflow: hidden;')
        
        self.menu.on_event('input', self.__internal_input)


    # Utility function to setup the button properties. Returns children and the icon flag
    def __setup_btn_children(self):
        children = []
        leftspace = 0
        flagicon = not self.icon is None
        if len(self.buttontext) > 0:
            children.append(self.buttontext)
            leftspace = 2
            flagicon = False
            
        if not self.icon is None:
            children.append(v.Icon(small=self.icon_small, large=self.icon_large, children=[self.icon], class_='pa-0 ma-0 ml-%d'%leftspace))
            
        return children, flagicon
        

    # Manage event
    def __internal_input(self, widget=None, event=None, data=None):
        if not self.on_input is None:
            if data is True:
                self.on_input(True)
            else:
                self.on_input(False)
    
    
    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display"""
        return self.menu

    
    @property
    def text(self):
        """
        Get/Set the text of the popup button.
        
        Returns
        --------
        text : str
               current text displayed on the popup button

        Example
        -------
        Change the text of the popup button::
            
            p.text = "New text"
            print(p.text)
        
        """
        return self.buttontext

    
    @text.setter
    def text(self, t):
        if isinstance(t, str):
            self.buttontext = t
            self.btn.children, self.btn.icon = self.__setup_btn_children()
            
            
    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state of the widget.
        
        Returns
        --------
        flag : bool
            True if the widget is disabled, False otherwise

        Example
        -------
        Programmatically change the disabled state::
            
            p.disabled = True
            print(p.disabled)
        
        """
        return self.btn.disabled

    
    @disabled.setter
    def disabled(self, flag):
        self.btn.disabled = flag
            
            
    # color property
    @property
    def color(self):
        """
        Get/Set the color of the widget.
        
        Returns
        --------
        c : str
            widget color in the hex format ('#xxxxxx')

        Example
        -------
        Programmatically change the widget color::
            
            p.color = '#00FF00'
            print(p.color)
        
        """
        return self.btn.color

    @color.setter
    def color(self, c):
        if isinstance(c, str):
            self.btn.color = c



"""Widget to upload files from the user local machine"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from ipywidgets import widgets, Layout
from IPython.display import display, HTML

try:
    from . import settings
except:
    import settings

from vois.vuetify.extra import FileInput


#####################################################################################################################################################
# Upload file input
#####################################################################################################################################################
class upload():
    """
    Widget to upload files from the user local machine.
        
    Parameters
    ----------
    accept : str, optional
        String containing the comma separated list of mime types of files accepted for the upload operation (example: 'image/png, image/jpeg')
    label : str, optional
        Label displayed in the upper part of the widget (default is the empty string)
    placeholder : str, optional
        String that provides some guidance to the user (default is the empty string)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    width : str, optional
        Width of the control in pixels or in percentage (default is "100%")
    margins : str, optional
        Margins to apply to the widgets (default is "pa-0 ma-0")
    multiple : bool, optional
        Flag to enable multiple selection of files to upload (default is False)
    show_progress : bool, optional
        Flag to show a progress bar while uploading (default is True)
    onchange : function, optional
        Python function to call when the user has selected  one or more files to upload. The function will receive a parameter of type list containing the files to upload
    onchanging : function, optional
        Python function to call just after the user selects one or more files to upload. The function will receive no parameters
            
    Example
    -------
    Creation and display of a widget for the upload of images::
        
        from vois.vuetify import upload
        from IPython.display import display
        
        u = upload.upload(accept="image/png, image/jpeg, image/bmp",
                          label='Images',
                          placeholder='Click to select images to upload')
        display(u.draw())


    .. figure:: figures/upload.png
       :scale: 100 %
       :alt: upload widget

       Upload widget for selecting images to upload.
   """
    
    # Initialization
    def __init__(self, accept='', label='', placeholder='', color=settings.color_first,
                 width="100%", margins="pa-0 ma-0",
                 multiple=False, show_progress=True, onchange=None, onchanging=None):

        self._color     = color
        self.width      = width
        self.margins    = margins
        self.onchange   = onchange
        self.onchanging = onchanging
        
        self.file_input = FileInput(accept=accept, color=self._color, multiple=multiple, show_progress=show_progress, 
                                    label=label, placeholder=placeholder)
        
        self.container = v.Card(flat=True, style_="min-width: %s; width: %s;"%(self.width,self.width), class_=self.margins, children=[self.file_input])
        
        # If requested onchange management
        if not self.onchange is None:
            self.file_input.observe(self.__internal_on_file_upload, names='file_info')
        
        
    # Clear
    def clear(self):
        """Sets the widget to its initial state (no file selected)"""
        self.file_input.clear()
    
        
    # Manage onchange event
    def __internal_on_file_upload(self, change):
        if self.onchanging:
            self.onchanging()
            
        files = self.file_input.get_files()
        
        if self.onchange:
            self.onchange(files)
    
    # Returns the vuetify object to display
    def draw(self):
        """
        Returns the ipyvuetify object to display
        
        Example
        -------
        Display the upload widget::
        
            import upload
            u = upload.upload(accept="image/png, image/jpeg, image/bmp",
                              label='Images',
                              placeholder='Click to select images to upload')
            display(u.draw())
        """
        return v.Html(tag='div',children=[self.container], style_='overflow: hidden;')


    @property
    def color(self):
        """
        Get/Set the color of the upload widget
        
        Returns
        --------
        col : str
            Color of the upload widget

        Example
        -------
        Programmatically change the color::
            
            p.color = 'red'
            print(p.color)
        
        """
        return self._color
        
    @color.setter
    def color(self, col):
        self._color = col
        self.file_input.color = self._color
    


"""SVG visualization of a graph."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import HTML, widgets, Layout
from ipyevents import Event
from IPython.display import display
import math
from textwrap import wrap

try:
    from . import colors
    from .vuetify import fontsettings
    from .vuetify import settings
except:
    import colors
    from vuetify import fontsettings
    from vuetify import settings


###########################################################################################################################################################################
# Visualization of a Graph with predefined positions
###########################################################################################################################################################################
def svgGraph(nodes_name,
             nodes_label,
             nodes_color,
             nodes_pos,
             edges,
             edge_label='Value',
             textcolor='white',
             edgestrokecolor='lightgrey',
             edgestrokewidth=0.3,
             nodestrokecolor='grey',
             nodestrokewidth=0.3,
             selectedcolor='red',
             selectedstrokewidth=0.7,
             width=300.0,
             heightpercent=100.0,
             borderspercent=10.0,
             nodesradius=3.0,
             fontsize=3.0,
             onclick=None):
    """
    Display of a graph.
    
    Parameters
    ----------
    nodes_name : dict
        Dictionary with Key=nodeid and Value=name (short label to display inside the node)
    nodes_label : dict
        Dictionary with Key=nodeid and Value=description (long description to display as tooltip of the nodes)
    nodes_color : dict
        Dictionary with Key=nodeid and Value=color to use for the node
    nodes_pos : dict
        Dictionary with Key=nodeid and Value=[x,y] coordinates of the node
    edges : dict
        Dictionary with Key=(nodeid1,nodeid2) and Value containing a numerical value to display in the tooltip
    edge_label : str, optional
        String to display inside the tooltip over an edge of the graph (default is 'Value')
    textcolor : str, optional
        Color to use for text of nodes (default is 'white')
    edgestrokecolor : str, optional
        Color to use to display the edges (default is 'lightgrey')
    edgestrokewidth : float, optional
        Stroke width of the line used to display the edges (default is 0.3)
    nodestrokecolor : str, optional
        Color to use for the border of the nodes (default is 'grey')
    nodestrokewidth : float, optional
        Stroke width of the line used to display the border of the nodes (default is 0.3)
    selectedcolor : str, optional
        Color to use to display the border of the selected node (default is 'red')
    selectedstrokewidth : float, optional
        Stroke width of the line used to display the border of the selected node (default is 0.7)
    width : int or float or str, optional
        Width of the drawing. If an integer or a float is passed, the size is intended in pixels units, otherwise a string containing the units must be passed (example: '4vw'). The default is 300.0 for 300 pixels
    heightpercent : float, optional
        Height of the drawing in percentage of the Width (default is 100.0, meaning that a square chart will be created)
    borderspercent : float, optional
        Border to add in percentage on each side of the chart to be sure that the nodes circles are completely inside (default is 10.0)
    nodesradius : float, optional
        Radius of the circles that represent the nodes. The total graph drawing is created in SVG coordinates [0,100]. Default is 3.0.
    fontsize : float, optional
        Size of the font used for nodes' texts. The total graph drawing is created in SVG coordinates [0,100]. Default is 3.0.
    onclick : function, optional
        Python function to call when the user clicks on one of the nodes of the graph. The function will receive as parameter the nodeid of the clicked node, or -1 when the click is outside of all the nodes
        
    Returns
    -------
        an ipywidgets.Output instance with the graph displayed inside
        
    Example
    -------
    Example of a generation and display of a random graph using networkx library and svgGraph module::
    
        from vois import svgGraph
        import networkx as nx
        import random
        r = lambda: random.randint(0,255)

        # Generate a random graph using networkx
        G = nx.gnp_random_graph(20, 0.2, seed=12345)
        
        # Assign position to nodes using Fruchterman-Reingold force-directed algorithm
        nodes_pos = nx.spring_layout(G, weight='weight', seed=12345)

        # Extract nodes and edges information from the graph
        nodes = list(G.nodes)
        edges = list(G.edges)
        
        # Generate dictionaries requested by the svgGraph module
        nodes_name  = dict(zip(nodes,['N'+str(x) for x in nodes]))
        nodes_label = dict(zip(nodes,['Description of node '+str(x) for x in nodes]))
        nodes_color = dict(zip(nodes,['#%02X%02X%02X' % (r(),r(),r()) for x in nodes]))  # Random colors

        edges_value = dict(zip(edges,[random.randint(0,100) for x in edges]))

        # Display the graph
        svgGraph.svgGraph(nodes_name, nodes_label, nodes_color, nodes_pos, edges_value, width=800)

    .. figure:: figures/graph.png
       :scale: 100 %
       :alt: Graph example

       Example of a graph visualization
    
    """
    
    # Calculate bounding box
    xmin = ymin =  1.0e+300
    xmax = ymax = -1.0e+300
    for node, coords in nodes_pos.items():
        x = coords[0]
        y = coords[1]
        if x < xmin: xmin = x
        if x > xmax: xmax = x
        if y < ymin: ymin = y
        if y > ymax: ymax = y
    
    dx = xmax - xmin
    dy = ymax - ymin
    
    # If the height is bigger than requested
    opimaldy = dx * heightpercent / 100.0
    if dy != opimaldy:
        ymin =  1.0e+300
        ymax = -1.0e+300
        f = opimaldy/dy
        for node, coords in nodes_pos.items():
            coords[1] *= f
            if coords[1] < ymin: ymin = coords[1]
            if coords[1] > ymax: ymax = coords[1]
        

    # Enlarge on each side
    enlarge = borderspercent / 100.0
    xmin -= enlarge*dx
    xmax += enlarge*dx
    ymin -= enlarge*dy
    ymax += enlarge*dy
    
    dx = xmax - xmin
    dy = ymax - ymin
    height = 100.0 * dy / dx
    
    radiussvg = nodesradius   # Circles radius in SVG coordinates

    # Convert coordinates to [0,100] range
    def x2svg(x):
        return 100.0 * (x-xmin)/(xmax-xmin)
    
    def y2svg(y):
        return height * (y-ymin)/(ymax-ymin)
        
    # Convert svg coordinates to nodes positions
    def svg2x(x):
        return xmin + x * (xmax-xmin) / 100.0
    
    def svg2y(y):
        return ymin + y * (ymax-ymin) / height
        
    # ID of the selected node    
    selected_node = -1
    
    def createSVG():        
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" viewBox="0 0 100 %d" xml:space="preserve">' % int(height)

        # CSS styling
        svg += '''
            <style type="text/css">
               @import url('%s');
            </style>
        ''' % (fontsettings.font_url)

        for edge, value in edges.items():
            n1 = edge[0]
            n2 = edge[1]
            if n1 in nodes_pos and n2 in nodes_pos:
                x1 = x2svg(nodes_pos[n1][0])
                y1 = y2svg(nodes_pos[n1][1])
                x2 = x2svg(nodes_pos[n2][0])
                y2 = y2svg(nodes_pos[n2][1])
                tooltip = nodes_label[n1] + '\n' + nodes_label[n2] + '\n%s: %f' % (edge_label,value)
                svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s;stroke-width:%f"><title>%s</title></line>' % (x1,y1, x2,y2, edgestrokecolor,edgestrokewidth, tooltip)


        for node, coords in nodes_pos.items():
            x = x2svg(coords[0])
            y = y2svg(coords[1])
            color = nodes_color[node]
            name  = nodes_name[node]
            label = nodes_label[node]

            sc = nodestrokecolor
            sw = nodestrokewidth
            if node == selected_node:
                sc = selectedcolor
                sw = selectedstrokewidth
                
            svg += '<circle cx="%f" cy="%f" fill="%s" r="%f" style="stroke:%s;stroke-width:%f"><title>%s</title></circle>' % (x,y,color,radiussvg,sc,sw,label)
            svg += '<text style="pointer-events: none" text-anchor="middle" x="%f" y="%f" font-size="%f" fill="%s" font-weight="500">%s</text>' % (x, y+fontsize/3, fontsize, textcolor, name)


        svg += '</svg>'
        return svg
    
    # Create an output widget and display SVG in it
    if isinstance(width, int):
        w = '%dpx' % width
    elif isinstance(width, float):
        w = '%fpx' % width
    else:
        w = str(width)
        
    out = widgets.Output(layout=Layout(width=w, height='calc(calc(%s * %f) + 18px)' % (w,0.01*height)))

    svg = createSVG()
    with out:
        display(HTML(svg))

    d = Event(source=out, watched_events=['click'])

    def handle_event(event):
        nonlocal selected_node
        
        x = event['relativeX']
        y = event['relativeY']
        w = event['boundingRectWidth']
        h = event['boundingRectHeight']
        xsvg = (x / w) * 100.0
        ysvg = (y / h) * height
        
        xnode = svg2x(xsvg)
        ynode = svg2y(ysvg)
        
        r = svg2x(50.0+radiussvg) - svg2x(50.0)
        
        new_selected_node = -1
        for node, coords in nodes_pos.items():
            x = coords[0]
            y = coords[1]
            if math.hypot(xnode - x, ynode - y) <= r:
                if node == selected_node:
                    new_selected_node = -1
                else:
                    new_selected_node = node

        if new_selected_node != selected_node:
            selected_node = new_selected_node
            if not onclick is None:
                onclick(selected_node)
            out.clear_output(wait=True)
            with out:
                display(HTML(createSVG()))
        

    d.on_dom_event(handle_event)
        
    return out



"""General settings for ipyvuetify widgets."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

#########################################################################################################################################
# Graphical settings for vuetify widgets
#########################################################################################################################################

# Main settings
dark_mode = False            # If False then the foreground color is 'black', otherwise it is 'white'
"""
Flag to control the dark_mode of all the vuetify widgets.

If this flag is False, the foreground color used for text is 'black, otherwise it is 'white'
"""


button_rounded = True             # If True the buttons have rounded borders
"""
Flag to control the appearance of all button widgets (also in the toggle module or inside the title and footer bars).

If True, the buttons will have rounded borders, if False, the buttons will have squared borders. 

The :py:class:`button.button` class has an optional parameter **rounded** that can be used to force a single button to be rounded or squared. This setting will have influence on all the buttons created without expliciting setting the **rounded** parameter. 

Example of changing the appearance of all buttons to be rounded::

    from vois.vuetify import settings, button
    settings.button_rounded = True

    import importlib
    importlib.reload(button)

    b = button.button('Round button', width=200, selected=True)
    display(b.draw())

.. figure:: figures/RoundButton.png
   :scale: 100 %
   :alt: RoundButton

   Example of a round button


Example of changing the appearance of all buttons to be squared::

    from vois.vuetify import settings, button
    settings.button_rounded = False

    import importlib
    importlib.reload(button)

    b = button.button('Squared button', width=200, selected=True)
    display(b.draw())

.. figure:: figures/SquaredButton.png
   :scale: 100 %
   :alt: RoundButton

   Example of a squared button

.. note::

    The usage of **importlib.reload** is needed to reload the button module after the settings.button_rounded flag is changed

"""


color_first = '#f8bd1a'        # Primary color (for instance used for selected buttons)
"""
Primary color to be used for all the widgets of the vuetify package (for instance the buttons that have the selected state True will have this color as background color)

"""


color_second = '#efefef'        # Secondary color (for instance used for buttons not selected)
"""
Secondary color to be used for all the widgets of the vuetify package (for instance the buttons that have the selected state False will have this color as background color)
"""


textcolor_dark       = 'white'    # Text color when the background is dark
textcolor_notdark    = 'black'    # Text color when the background is not dark

tooltip_backcolor    = '#999999'  # Back color for tooltips


dark_background      = '#1e1e1e'  # Back color when dark mode is True

# Colors YELLOW
#color_first  = '#f8bd1a'     # Primary color (for instance used for selected buttons)
#color_second = '#efefef'     # Secondary color (for instance used for buttons not selected)

# Colors BLU
#color_first  = '#2196f3'     # Primary color (for instance used for selected buttons)
#color_second = '#95cffb'     # Secondary color (for instance used for buttons not selected)

# Colors GREEN
#color_first  = '#b7ec1c'     # Primary color (for instance used for selected buttons)
#color_second = '#9ebf00'     # Secondary color (for instance used for buttons not selected)

# Colors BLU EUROPEAN COMMISSION
#color_first  = '#254aa5'     # Primary color (for instance used for selected buttons)
#color_second = '#254aa5'     # Secondary color (for instance used for buttons not selected)


# Other derived settings
select_textcolor = textcolor_notdark

# Color for text on labels
label_textcolor  = select_textcolor

# Color of lines
line_color = select_textcolor


